# pylint: disable=protected-access
"""
Tests the functions from the RiskEstimators class.
"""

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.portfolio_optimization.risk_estimators import RiskEstimators
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators


class TestRiskEstimators(unittest.TestCase):
    """
    Tests different functions of the RiskEstimators class.
    """

    def setUp(self):
        """
        Initialize and get the test data
        """

        # Stock prices data to test the Covariance functions
        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date")

        # And series of returns
        ret_est = ReturnsEstimators()
        self.returns = ret_est.calculate_returns(self.data)

    def test_mp_pdf(self):
        """
        Test the deriving of pdf of the Marcenko-Pastur distribution.
        """

        risk_estimators = RiskEstimators()

        # Properties for the distribution
        var = 0.1
        tn_relation = 5
        num_points = 5

        # Calculating the pdf in 5 points
        pdf_mp = risk_estimators._mp_pdf(var, tn_relation, num_points)

        # Testing the minimum and maximum and non-zero values of the pdf
        self.assertAlmostEqual(pdf_mp.index[0], 0.03056, delta=1e-4)
        self.assertAlmostEqual(pdf_mp.index[4], 0.20944, delta=1e-4)

        # Testing that the distribution curve is right
        self.assertTrue(pdf_mp.values[1] > pdf_mp.values[2] > pdf_mp.values[3])

    def test_fit_kde(self):
        """
        Test the kernel fitting to a series of observations.
        """

        risk_estimators = RiskEstimators()

        # Values to fit kernel to and evaluation points
        observations = np.array([0.1, 0.2, 0.2, 0.3, 0.3, 0.3, 0.4, 0.4, 0.5])
        eval_points = np.array([0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6])

        # Calculating the pdf in 7 chosen points
        pdf_kde = risk_estimators._fit_kde(observations, eval_points=eval_points, kde_bwidth=0.25)

        # Testing the values and if the pdf is symmetric
        self.assertEqual(pdf_kde[0.0], pdf_kde[0.6])
        self.assertEqual(pdf_kde[0.1], pdf_kde[0.5])
        self.assertEqual(pdf_kde[0.2], pdf_kde[0.4])
        self.assertAlmostEqual(pdf_kde[0.3], 1.44413, delta=1e-5)

        # Testing also on unique values of the set as a default output
        pdf_kde_default = risk_estimators._fit_kde(observations, kde_bwidth=0.25)
        self.assertEqual(pdf_kde[0.1], pdf_kde_default[0.1])
        self.assertEqual(pdf_kde_default[0.2], pdf_kde_default[0.4])

    def test_pdf_fit(self):
        """
        Test the fit between empirical pdf and the theoretical pdf.
        """

        risk_estimators = RiskEstimators()

        # Values to calculate theoretical and empirical pdfs
        var = 0.6
        eigen_observations = np.array([0.1, 0.2, 0.2, 0.3, 0.3, 0.3, 0.4, 0.4, 0.5])
        tn_relation = 2
        kde_bwidth = 0.4

        # Calculating the SSE
        pdf_kde = risk_estimators._pdf_fit(var, eigen_observations, tn_relation, kde_bwidth)

        # Testing the SSE value
        self.assertAlmostEqual(pdf_kde, 50.51326, delta=1e-5)

    def test_find_max_eval(self):
        """
        Test the search for maximum random eigenvalue.
        """

        risk_estimators = RiskEstimators()

        # Values to calculate theoretical and empirical pdfs
        eigen_observations = np.array([0.1, 0.2, 0.2, 0.3, 0.3, 0.3, 0.4, 0.4, 0.5])
        tn_relation = 2
        kde_bwidth = 0.4

        # Optimizing and getting the maximum random eigenvalue and the optimal variation
        maximum_eigen, var = risk_estimators._find_max_eval(eigen_observations, tn_relation, kde_bwidth)

        # Testing the maximum random eigenvalue and the optimal variation
        self.assertAlmostEqual(maximum_eigen, 2.41011, delta=1e-5)
        self.assertAlmostEqual(var, 0.82702, delta=1e-5)

    @staticmethod
    def test_corr_to_cov():
        """
        Test the recovering of the covariance matrix from the correlation matrix.
        """

        risk_estimators = RiskEstimators()

        # Correlation matrix and the vector of standard deviations
        corr_matrix = np.array([[1, 0.1, -0.1],
                                [0.1, 1, -0.3],
                                [-0.1, -0.3, 1]])
        std_vec = np.array([0.1, 0.2, 0.1])

        # Expected covariance matrix
        expected_matrix = np.array([[0.01, 0.002, -0.001],
                                    [0.002, 0.04, -0.006],
                                    [-0.001, -0.006, 0.01]])

        # Finding the covariance matrix
        cov_matrix = risk_estimators.corr_to_cov(corr_matrix, std_vec)

        # Testing the first row of the matrix
        np.testing.assert_almost_equal(cov_matrix, expected_matrix, decimal=5)

    @staticmethod
    def test_cov_to_corr():
        """
        Test the deriving of the correlation matrix from a covariance matrix.
        """

        risk_estimators = RiskEstimators()

        # Covariance matrix
        cov_matrix = np.array([[0.01, 0.002, -0.001],
                               [0.002, 0.04, -0.006],
                               [-0.001, -0.006, 0.01]])

        # Expected correlation matrix
        expected_matrix = np.array([[1, 0.1, -0.1],
                                    [0.1, 1, -0.3],
                                    [-0.1, -0.3, 1]])

        # Finding the covariance matrix
        corr_matrix = risk_estimators.cov_to_corr(cov_matrix)

        # Testing the first row of the matrix
        np.testing.assert_almost_equal(corr_matrix, expected_matrix, decimal=5)

    @staticmethod
    def test_get_pca():
        """
        Test the calculation of eigenvalues and eigenvectors from a Hermitian matrix.
        """

        risk_estimators = RiskEstimators()

        # Correlation matrix as an input
        corr_matrix = np.array([[1, 0.1, -0.1],
                                [0.1, 1, -0.3],
                                [-0.1, -0.3, 1]])

        # Expected correlation matrix
        expected_eigenvalues = np.array([[1.3562, 0, 0],
                                         [0, 0.9438, 0],
                                         [0, 0, 0.7]])
        first_eigenvector = np.array([-3.69048184e-01, -9.29410263e-01, 1.10397126e-16])

        # Finding the eigenvalues
        eigenvalues, eigenvectors = risk_estimators._get_pca(corr_matrix)

        # Testing eigenvalues and the first eigenvector
        np.testing.assert_almost_equal(eigenvalues, expected_eigenvalues, decimal=4)
        np.testing.assert_almost_equal(eigenvectors[0], first_eigenvector, decimal=5)

    @staticmethod
    def test_denoised_corr():
        """
        Test the shrinkage the eigenvalues associated with noise.
        """

        risk_estimators = RiskEstimators()

        # Eigenvalues and eigenvectors to use
        eigenvalues = np.array([[1.3562, 0, 0],
                                [0, 0.9438, 0],
                                [0, 0, 0.7]])
        eigenvectors = np.array([[-3.69048184e-01, -9.29410263e-01, 1.10397126e-16],
                                 [-6.57192300e-01, 2.60956474e-01, 7.07106781e-01],
                                 [6.57192300e-01, -2.60956474e-01, 7.07106781e-01]])

        # Expected correlation matrix
        expected_corr = np.array([[1, 0.13353165, -0.13353165],
                                  [0.13353165, 1, -0.21921986],
                                  [-0.13353165, -0.21921986, 1]])

        # Finding the de-noised correlation matrix
        corr_matrix = risk_estimators._denoised_corr(eigenvalues, eigenvectors, 1)

        # Testing if the de-noised correlation matrix is right
        np.testing.assert_almost_equal(corr_matrix, expected_corr, decimal=4)

    @staticmethod
    def test_denoised_corr_targ_shrink():
        """
        Test the second method of shrinkage of the eigenvalues associated with noise.
        """

        risk_estimators = RiskEstimators()

        # Eigenvalues and eigenvectors to use
        eigenvalues = np.array([[1.3562, 0, 0],
                                [0, 0.9438, 0],
                                [0, 0, 0.7]])
        eigenvectors = np.array([[-3.69048184e-01, -9.29410263e-01, 1.10397126e-16],
                                 [-6.57192300e-01, 2.60956474e-01, 7.07106781e-01],
                                 [6.57192300e-01, -2.60956474e-01, 7.07106781e-01]])

        # Expected correlation matrix
        expected_corr = np.array([[1, 0.32892949, -0.32892949],
                                  [0.32892949, 1, -0.58573558],
                                  [-0.32892949, -0.58573558, 1]])

        # Finding the de-noised correlation matrix
        corr_matrix = risk_estimators._denoised_corr_targ_shrink(eigenvalues, eigenvectors, 1)

        # Testing if the de-noised correlation matrix is right
        np.testing.assert_almost_equal(corr_matrix, expected_corr, decimal=4)

    @staticmethod
    def test_detoned():
        """
        Test the de-toning of the correlation matrix.
        """

        risk_estimators = RiskEstimators()

        # Correlation matrix to use
        corr = np.array([[1, 0.1, -0.1],
                         [0.1, 1, -0.3],
                         [-0.1, -0.3, 1]])

        # Eigenvalues and eigenvectors to use
        eigenvalues = np.array([[1.3562, 0, 0],
                                [0, 0.9438, 0],
                                [0, 0, 0.7]])
        eigenvectors = np.array([[-3.69048184e-01, -9.29410263e-01, 1.10397126e-16],
                                 [-6.57192300e-01, 2.60956474e-01, 7.07106781e-01],
                                 [6.57192300e-01, -2.60956474e-01, 7.07106781e-01]])

        # Expected correlation matrix
        expected_corr = np.array([[1, -0.33622026, 0.33622026],
                                  [-0.33622026, 1, 0.88478197],
                                  [0.33622026, 0.88478197, 1]])

        # Finding the de-toned correlation matrix
        corr_matrix = risk_estimators._detoned_corr(corr, eigenvalues, eigenvectors, 1)

        # Testing if the de-toned correlation matrix is right
        np.testing.assert_almost_equal(corr_matrix, expected_corr, decimal=4)

    @staticmethod
    def test_denoise_covariance():
        """
        Test the shrinkage the eigenvalues associated with noise.
        """

        risk_estimators = RiskEstimators()

        # Covariance matrix to de-noise and parameters for the theoretical distribution.
        cov_matrix = np.array([[0.01, 0.002, -0.001],
                               [0.002, 0.04, -0.006],
                               [-0.001, -0.006, 0.01]])
        tn_relation = 50
        kde_bwidth = 0.25
        alpha = 0.2
        denoise_method = 'const_resid_eigen'
        denoise_method_alt = 'target_shrink'
        detone = False
        detone_alt = True
        market_component = 1

        # Expected de-noised covariance matrix
        expected_cov = np.array([[0.01, 0.00267029, -0.00133514],
                                 [0.00267029, 0.04, -0.00438387],
                                 [-0.00133514, -0.00438387, 0.01]])

        expected_cov_alt = np.array([[0.01, 0.0057, -0.0028],
                                     [0.0057, 0.04, -0.0106],
                                     [-0.0028, -0.0106, 0.01]])

        expected_cov_detoned = np.array([[0.01, -0.00672445, 0.00336222],
                                         [-0.00672445, 0.04, 0.01769514],
                                         [0.00336222, 0.01769514, 0.01]])

        # Finding the de-noised covariance matrix
        cov_matrix_denoised = risk_estimators.denoise_covariance(cov_matrix, tn_relation, denoise_method, detone,
                                                                 market_component, kde_bwidth)

        # Using the alternative de-noising method
        cov_matrix_denoised_alt = risk_estimators.denoise_covariance(cov_matrix, tn_relation, denoise_method_alt,
                                                                     detone, market_component, kde_bwidth, alpha)

        # Finding the de-toned covariance matrix
        cov_matrix_detoned = risk_estimators.denoise_covariance(cov_matrix, tn_relation, denoise_method, detone_alt,
                                                                market_component, kde_bwidth)

        # Testing if the de-noised covariance matrix is right
        np.testing.assert_almost_equal(cov_matrix_denoised, expected_cov, decimal=4)

        # Testing if the de-noised covariance matrix is right
        np.testing.assert_almost_equal(cov_matrix_denoised_alt, expected_cov_alt, decimal=4)

        # Testing if the de-toned covariance matrix is right
        np.testing.assert_almost_equal(cov_matrix_detoned, expected_cov_detoned, decimal=4)

    def test_minimum_covariance_determinant(self):
        """
        Test the calculation of the Minimum Covariance Determinant.
        """

        risk_estimators = RiskEstimators()

        # Getting first three columns of data to be able to compare the output
        prices_dataframe = self.data.iloc[:, :3]
        returns_dataframe = self.returns.iloc[:, :3]

        # Expected resulting Minimum Covariance Determinant
        expected_cov = np.array([[1.5110e-04, 1.1322e-04, -5.2053e-06],
                                 [1.1322e-04, 1.4760e-06, -6.6961e-06],
                                 [-5.2053e-06, -6.6961e-06, 1.0874e-05]])

        # Using the Minimum Covariance Determinant algorithm on price data with random seed 0
        min_covar_determ = risk_estimators.minimum_covariance_determinant(prices_dataframe, price_data=True,
                                                                          random_state=0)

        # Using the Minimum Covariance Determinant algorithm on return data with random seed 0
        min_covar_determ_ret = risk_estimators.minimum_covariance_determinant(returns_dataframe, price_data=False,
                                                                              random_state=0)

        # Testing if the resulting covariance matrix is right
        np.testing.assert_almost_equal(min_covar_determ, expected_cov, decimal=4)

        # And if the results for price and returns are the same
        np.testing.assert_almost_equal(min_covar_determ, min_covar_determ_ret, decimal=4)

    def test_empirical_covariance(self):
        """
        Test the calculation of the Maximum likelihood covariance estimator.
        """

        risk_estimators = RiskEstimators()

        # Getting first three columns of data to be able to compare the output
        prices_dataframe = self.data.iloc[:, :3]
        returns_dataframe = self.returns.iloc[:, :3]

        # Expected resulting Maximum likelihood covariance estimator
        expected_cov = np.array([[4.6571e-04, 3.4963e-04, -1.6626e-05],
                                 [3.4963e-04, 3.7193e-04, -1.4957e-05],
                                 [-1.6626e-05, -1.4957e-05, 1.9237e-05]])

        # Using the Maximum likelihood covariance estimator on price data
        empirical_cov = risk_estimators.empirical_covariance(prices_dataframe, price_data=True)

        # Using the Maximum likelihood covariance estimator on returns data
        empirical_cov_ret = risk_estimators.empirical_covariance(returns_dataframe, price_data=False)

        # Testing if the resulting covariance matrix is right
        np.testing.assert_almost_equal(empirical_cov, expected_cov, decimal=6)

        # And if the results for price and returns are the same
        np.testing.assert_almost_equal(empirical_cov, empirical_cov_ret, decimal=4)

    def test_shrinked_covariance(self):
        """
        Test the calculation of the Covariance estimator with shrinkage.
        """

        risk_estimators = RiskEstimators()

        # Getting first three columns of data to be able to compare the output
        prices_dataframe = self.data.iloc[:, :3]
        returns_dataframe = self.returns.iloc[:, :3]

        # Expected resulting Covariance estimators for each shrinkage type
        expected_cov_basic = np.array([[4.47705356e-04, 3.14668132e-04, -1.49635474e-05],
                                       [3.14668132e-04, 3.63299625e-04, -1.34611717e-05],
                                       [-1.49635474e-05, -1.34611717e-05, 4.58764444e-05]])

        expected_cov_lw = np.array([[4.63253312e-04, 3.44853842e-04, -1.63989814e-05],
                                    [3.44853842e-04, 3.70750646e-04, -1.47524847e-05],
                                    [-1.63989814e-05, -1.47524847e-05, 2.28774674e-05]])

        expected_cov_oas = np.array([[4.65398835e-04, 3.49019287e-04, -1.65970625e-05],
                                     [3.49019287e-04, 3.71778842e-04, -1.49306780e-05],
                                     [-1.65970625e-05, -1.49306780e-05, 1.97037481e-05]])

        # Using the Covariance estimator with different types of shrinkage on price data
        shrinked_cov_basic = risk_estimators.shrinked_covariance(prices_dataframe, price_data=True,
                                                                 shrinkage_type='basic', basic_shrinkage=0.1)

        shrinked_cov_lw = risk_estimators.shrinked_covariance(prices_dataframe, price_data=True, shrinkage_type='lw')

        shrinked_cov_oas = risk_estimators.shrinked_covariance(prices_dataframe, price_data=True, shrinkage_type='oas')

        shrinked_cov_all = risk_estimators.shrinked_covariance(prices_dataframe, price_data=True,
                                                               shrinkage_type='all', basic_shrinkage=0.1)

        # Using the Covariance estimator with different types of shrinkage on returns data
        shrinked_cov_basic_ret = risk_estimators.shrinked_covariance(returns_dataframe, price_data=False,
                                                                     shrinkage_type='basic', basic_shrinkage=0.1)

        # Testing if the resulting shrinked covariance matrix is right for every method is right
        np.testing.assert_almost_equal(shrinked_cov_basic, expected_cov_basic, decimal=7)
        np.testing.assert_almost_equal(shrinked_cov_lw, expected_cov_lw, decimal=7)
        np.testing.assert_almost_equal(shrinked_cov_oas, expected_cov_oas, decimal=7)

        # And that the results from all methods match the individual methods results
        np.testing.assert_almost_equal(shrinked_cov_all[0], shrinked_cov_basic, decimal=7)
        np.testing.assert_almost_equal(shrinked_cov_all[1], shrinked_cov_lw, decimal=7)
        np.testing.assert_almost_equal(shrinked_cov_all[2], shrinked_cov_oas, decimal=7)

        # And if the results for price and returns are the same
        np.testing.assert_almost_equal(shrinked_cov_basic, shrinked_cov_basic_ret, decimal=4)

    def test_semi_covariance(self):
        """
        Test the calculation of the Semi-Covariance matrix.
        """

        risk_estimators = RiskEstimators()

        # Getting first three columns of data to be able to compare the output
        prices_dataframe = self.data.iloc[:, :3]
        returns_dataframe = self.returns.iloc[:, :3]

        # Expected Semi-Covariance matrix
        expected_semi_cov = np.array([[7.302402e-05, 5.855724e-05, 3.075326e-06],
                                      [5.855724e-05, 6.285548e-05, 2.788988e-06],
                                      [3.075326e-06, 2.788988e-06, 3.221170e-06]])

        # Calculating the Semi-Covariance matrix on price data with zero threshold (volatility of negative returns)
        semi_cov = risk_estimators.semi_covariance(prices_dataframe, price_data=True, threshold_return=0)

        # Calculating the Semi-Covariance matrix on returns data with zero threshold (volatility of negative returns)
        semi_cov_ret = risk_estimators.semi_covariance(returns_dataframe, price_data=False, threshold_return=0)

        # Testing if the resulting Semi-Covariance matrix is right
        np.testing.assert_almost_equal(semi_cov, expected_semi_cov, decimal=6)

        # And if the results for price and returns are the same
        np.testing.assert_almost_equal(np.array(semi_cov), np.array(semi_cov_ret), decimal=4)

    def test_exponential_covariance(self):
        """
        Test the calculation of the Exponentially-weighted Covariance matrix.
        """

        risk_estimators = RiskEstimators()

        # Getting first three columns of data to be able to compare the output
        prices_dataframe = self.data.iloc[:, :3]
        returns_dataframe = self.returns.iloc[:, :3]

        # Expected Exponentially-weighted Covariance matrix
        expected_expon_cov = np.array([[2.824303e-04, 3.215506e-04, -4.171518e-06],
                                       [3.215506e-04, 4.585646e-04, -1.868617e-05],
                                       [-4.171518e-06, -1.868617e-05, 8.684991e-06]])

        # Calculating the Exponentially-weighted Covariance matrix on price data with the span of 60
        expon_cov = risk_estimators.exponential_covariance(prices_dataframe, price_data=True, window_span=60)

        # Calculating the Exponentially-weighted Covariance matrix on price data with the span of 60
        expon_cov_ret = risk_estimators.exponential_covariance(returns_dataframe, price_data=False, window_span=60)

        # Testing if the resulting Semi-Covariance matrix is right
        np.testing.assert_almost_equal(expon_cov, expected_expon_cov, decimal=6)

        # And if the results for price and returns are the same
        np.testing.assert_almost_equal(np.array(expon_cov), np.array(expon_cov_ret), decimal=4)



"""
Correlation based distances and various modifications (angular, absolute, squared) described in Cornell lecture notes:
Codependence: https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes
"""

import numpy as np
from scipy.spatial.distance import squareform, pdist


# pylint: disable=invalid-name


def angular_distance(x: np.array, y: np.array) -> float:
    """
    Returns angular distance between two vectors. Angular distance is a slight modification of Pearson correlation which
    satisfies metric conditions.

    Formula used for calculation:

    Ang_Distance = (1/2 * (1 - Corr))^(1/2)

    Read Cornell lecture notes for more information about angular distance:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes.

    :param x: (np.array/pd.Series) X vector.
    :param y: (np.array/pd.Series) Y vector.
    :return: (float) Angular distance.
    """

    corr_coef = np.corrcoef(x, y)[0][1]
    return np.sqrt(0.5 * (1 - corr_coef))


def absolute_angular_distance(x: np.array, y: np.array) -> float:
    """
    Returns absolute angular distance between two vectors. It is a modification of angular distance where the absolute
    value of the Pearson correlation coefficient is used.

    Formula used for calculation:

    Abs_Ang_Distance = (1/2 * (1 - abs(Corr)))^(1/2)

    Read Cornell lecture notes for more information about absolute angular distance:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes.

    :param x: (np.array/pd.Series) X vector.
    :param y: (np.array/pd.Series) Y vector.
    :return: (float) Absolute angular distance.
    """

    corr_coef = np.corrcoef(x, y)[0][1]
    return np.sqrt(0.5 * (1 - abs(corr_coef)))


def squared_angular_distance(x: np.array, y: np.array) -> float:
    """
    Returns squared angular distance between two vectors. It is a modification of angular distance where the square of
    Pearson correlation coefficient is used.

    Formula used for calculation:

    Squared_Ang_Distance = (1/2 * (1 - (Corr)^2))^(1/2)

    Read Cornell lecture notes for more information about squared angular distance:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes.

    :param x: (np.array/pd.Series) X vector.
    :param y: (np.array/pd.Series) Y vector.
    :return: (float) Squared angular distance.
    """

    corr_coef = np.corrcoef(x, y)[0][1]
    return np.sqrt(0.5 * (1 - corr_coef ** 2))


def distance_correlation(x: np.array, y: np.array) -> float:
    """
    Returns distance correlation between two vectors. Distance correlation captures both linear and non-linear
    dependencies.

    Formula used for calculation:

    Distance_Corr[X, Y] = dCov[X, Y] / (dCov[X, X] * dCov[Y, Y])^(1/2)

    dCov[X, Y] is the average Hadamard product of the doubly-centered Euclidean distance matrices of X, Y.

    Read Cornell lecture notes for more information about distance correlation:
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3512994&download=yes.

    :param x: (np.array/pd.Series) X vector.
    :param y: (np.array/pd.Series) Y vector.
    :return: (float) Distance correlation coefficient.
    """

    x = x[:, None]
    y = y[:, None]

    x = np.atleast_2d(x)
    y = np.atleast_2d(y)

    a = squareform(pdist(x))
    b = squareform(pdist(y))

    A = a - a.mean(axis=0)[None, :] - a.mean(axis=1)[:, None] + a.mean()
    B = b - b.mean(axis=0)[None, :] - b.mean(axis=1)[:, None] + b.mean()

    d_cov_xx = (A * A).sum() / (x.shape[0] ** 2)
    d_cov_xy = (A * B).sum() / (x.shape[0] ** 2)
    d_cov_yy = (B * B).sum() / (x.shape[0] ** 2)

    coef = np.sqrt(d_cov_xy) / np.sqrt(np.sqrt(d_cov_xx) * np.sqrt(d_cov_yy))

    return coef



"""
Module which implements feature importance algorithms as described in Chapter 8 of Advances in Financial Machine
Learning and Clustered Feature Importance algorithms as described in Chapter 6 Section 6.5.2 of Machine Learning for
Asset Managers.
"""

import pandas as pd
import numpy as np
from sklearn.metrics import log_loss
import matplotlib.pyplot as plt

from famlafl.cross_validation.cross_validation import ml_cross_val_score


# pylint: disable=invalid-name
# pylint: disable=invalid-unary-operand-type
# pylint: disable=comparison-with-callable
# pylint: disable=too-many-locals

def mean_decrease_impurity(model, feature_names, clustered_subsets=None):
    """
    Advances in Financial Machine Learning, Snippet 8.2, page 115.

    MDI Feature importance

    Mean decrease impurity (MDI) is a fast, explanatory-importance (in-sample, IS) method specific to tree-based
    classifiers, like RF. At each node of each decision tree, the selected feature splits the subset it received in
    such a way that impurity is decreased. Therefore, we can derive for each decision tree how much of the overall
    impurity decrease can be assigned to each feature. And given that we have a forest of trees, we can average those
    values across all estimators and rank the features accordingly.

    Tip:
    Masking effects take place when some features are systematically ignored by tree-based classifiers in favor of
    others. In order to avoid them, set max_features=int(1) when using sklearn’s RF class. In this way, only one random
    feature is considered per level.

    Notes:

    * MDI cannot be generalized to other non-tree based classifiers
    * The procedure is obviously in-sample.
    * Every feature will have some importance, even if they have no predictive power whatsoever.
    * MDI has the nice property that feature importances add up to 1, and every feature importance is bounded between 0 and 1.
    * method does not address substitution effects in the presence of correlated features. MDI dilutes the importance of
      substitute features, because of their interchangeability: The importance of two identical features will be halved,
      as they are randomly chosen with equal probability.
    * Sklearn’s RandomForest class implements MDI as the default feature importance score. This choice is likely
      motivated by the ability to compute MDI on the fly, with minimum computational cost.

    Clustered Feature Importance( Machine Learning for Asset Manager snippet 6.4 page 86) :
    Clustered MDI  is the  modified version of MDI (Mean Decreased Impurity). It  is robust to substitution effect that
    takes place when two or more explanatory variables share a substantial amount of information (predictive power).CFI
    algorithm described by Dr Marcos Lopez de Prado  in Clustered Feature  Importance section of book Machine Learning
    for Asset Manager. Here  instead of  taking the importance  of  every feature, we consider the importance of every
    feature subsets, thus every feature receive the importance of subset it belongs to.

    :param model: (model object): Trained tree based classifier.
    :param feature_names: (list): Array of feature names.
    :param clustered_subsets: (list) Feature clusters for Clustered Feature Importance (CFI). Default None will not apply CFI.
                              Structure of the input must be a list of list/s i.e. a list containing the clusters/subsets of feature
                              name/s inside a list. E.g- [['I_0','I_1','R_0','R_1'],['N_1','N_2'],['R_3']]
    :return: (pd.DataFrame): Mean and standard deviation feature importance.
    """
    # Feature importance based on in-sample (IS) mean impurity reduction
    feature_imp_df = {i: tree.feature_importances_ for i, tree in enumerate(model.estimators_)}
    feature_imp_df = pd.DataFrame.from_dict(feature_imp_df, orient='index')
    feature_imp_df.columns = feature_names

    # Make sure that features with zero importance are not averaged, since the only reason for a 0 is that the feature
    # was not randomly chosen. Replace those values with np.nan
    feature_imp_df = feature_imp_df.replace(0, np.nan)  # Because max_features = 1

    if clustered_subsets is not None:
        # Getting subset wise importance
        importance = pd.DataFrame(index=feature_names, columns=['mean', 'std'])
        for subset in clustered_subsets: # Iterating over each cluster
            subset_feat_imp = feature_imp_df[subset].sum(axis=1)
            # Importance of each feature within a subsets is equal to the importance of that subset
            importance.loc[subset, 'mean'] = subset_feat_imp.mean()
            importance.loc[subset, 'std'] = subset_feat_imp.std()*subset_feat_imp.shape[0]**-.5
    else:
        importance = pd.concat({'mean': feature_imp_df.mean(),
                                'std': feature_imp_df.std() * feature_imp_df.shape[0] ** -0.5},
                               axis=1)

    importance /= importance['mean'].sum()
    return importance


def mean_decrease_accuracy(model, X, y, cv_gen, clustered_subsets=None, sample_weight_train=None,
                           sample_weight_score=None, scoring=log_loss, random_state=42):
    """
    Advances in Financial Machine Learning, Snippet 8.3, page 116-117.

    MDA Feature Importance

    Mean decrease accuracy (MDA) is a slow, predictive-importance (out-of-sample, OOS) method. First, it fits a
    classifier; second, it derives its performance OOS according to some performance score (accuracy, negative log-loss,
    etc.); third, it permutates each column of the features matrix (X), one column at a time, deriving the performance
    OOS after each column’s permutation. The importance of a feature is a function of the loss in performance caused by
    its column’s permutation. Some relevant considerations include:

    * This method can be applied to any classifier, not only tree-based classifiers.
    * MDA is not limited to accuracy as the sole performance score. For example, in the context of meta-labeling
      applications, we may prefer to score a classifier with F1 rather than accuracy. That is one reason a better
      descriptive name would have been “permutation importance.” When the scoring function does not correspond to a
      metric space, MDA results should be used as a ranking.
    * Like MDI, the procedure is also susceptible to substitution effects in the presence of correlated features.
      Given two identical features, MDA always considers one to be redundant to the other. Unfortunately, MDA will make
      both features appear to be outright irrelevant, even if they are critical.
    * Unlike MDI, it is possible that MDA concludes that all features are unimportant. That is because MDA is based on
      OOS performance.
    * The CV must be purged and embargoed.

    Clustered Feature Importance( Machine Learning for Asset Manager snippet 6.5 page 87) :
    Clustered MDA is the modified version of MDA (Mean Decreased Accuracy). It is robust to substitution effect that takes
    place when two or more explanatory variables share a substantial amount of information (predictive power).CFI algorithm
    described by Dr Marcos Lopez de Prado  in Clustered Feature  Importance (Presentation Slides)
    https://papers.ssrn.com/sol3/papers.cfm?abstract_id=3517595. Instead of shuffling (permutating) all variables
    individually (like in MDA), we shuffle all variables in cluster together. Next, we follow all the  rest of the
    steps as in MDA. It can used by simply specifying the clustered_subsets argument.

    :param model: (sklearn.Classifier): Any sklearn classifier.
    :param X: (pd.DataFrame): Train set features.
    :param y: (pd.DataFrame, np.array): Train set labels.
    :param cv_gen: (cross_validation.PurgedKFold): Cross-validation object.
    :param clustered_subsets: (list) Feature clusters for Clustered Feature Importance (CFI). Default None will not apply CFI.
                              Structure of the input must be a list of list/s i.e. a list containing the clusters/subsets of feature
                              name/s inside a list. E.g- [['I_0','I_1','R_0','R_1'],['N_1','N_2'],['R_3']]
    :param sample_weight_train: (np.array) Sample weights used to train the model for each record in the dataset.
    :param sample_weight_score: (np.array) Sample weights used to evaluate the model quality.
    :param scoring: (function): Scoring function used to determine importance.
    :param random_state: (int) Random seed for shuffling the features.
    :return: (pd.DataFrame): Mean and standard deviation of feature importance.
    """

    if sample_weight_train is None:
        sample_weight_train = np.ones((X.shape[0],))

    if sample_weight_score is None:
        sample_weight_score = np.ones((X.shape[0],))

    fold_metrics_values, features_metrics_values = pd.Series(dtype='float64'), pd.DataFrame(columns=X.columns)
    # Generating a numpy random state object for the given random_state
    rs_obj = np.random.RandomState(seed=random_state)
    # Clustered feature subsets will be used for CFI if clustered_subsets exists else will operate on the single column as MDA
    feature_sets = clustered_subsets if clustered_subsets else [[x] for x in X.columns]
    for i, (train, test) in enumerate(cv_gen.split(X=X)):
        fit = model.fit(X=X.iloc[train, :], y=y.iloc[train], sample_weight=sample_weight_train[train])
        pred = fit.predict(X.iloc[test, :])

        # Get overall metrics value on out-of-sample fold
        if scoring == log_loss:
            prob = fit.predict_proba(X.iloc[test, :])
            fold_metrics_values.loc[i] = -scoring(y.iloc[test], prob, sample_weight=sample_weight_score[test],
                                                  labels=model.classes_)
        else:
            fold_metrics_values.loc[i] = scoring(y.iloc[test], pred, sample_weight=sample_weight_score[test])

        # Get feature specific metric on out-of-sample fold
        for j in feature_sets:
            X1_ = X.iloc[test, :].copy(deep=True)
            for j_i in j:
                rs_obj.shuffle(X1_[j_i].values)  # Permutation of a single column for MDA or through the whole subset for CFI
            if scoring == log_loss:
                prob = fit.predict_proba(X1_)
                features_metrics_values.loc[i, j] = -scoring(y.iloc[test], prob,
                                                             sample_weight=sample_weight_score[test],
                                                             labels=model.classes_)
            else:
                pred = fit.predict(X1_)
                features_metrics_values.loc[i, j] = scoring(y.iloc[test], pred,
                                                            sample_weight=sample_weight_score[test])

    importance = (-features_metrics_values).add(fold_metrics_values, axis=0)
    if scoring == log_loss:
        importance = importance / -features_metrics_values
    else:
        importance = importance / (1.0 - features_metrics_values).replace(0, np.nan)
    importance = pd.concat({'mean': importance.mean(), 'std': importance.std() * importance.shape[0] ** -.5}, axis=1)
    importance.replace([-np.inf, np.nan], 0, inplace=True)  # Replace infinite values

    return importance


def single_feature_importance(clf, X, y, cv_gen, sample_weight_train=None, sample_weight_score=None, scoring=log_loss):
    """
    Advances in Financial Machine Learning, Snippet 8.4, page 118.

    Implementation of SFI

    Substitution effects can lead us to discard important features that happen to be redundant. This is not generally a
    problem in the context of prediction, but it could lead us to wrong conclusions when we are trying to understand,
    improve, or simplify a model. For this reason, the following single feature importance method can be a good
    complement to MDI and MDA.

    Single feature importance (SFI) is a cross-section predictive-importance (out-of- sample) method. It computes the
    OOS performance score of each feature in isolation. A few considerations:

    * This method can be applied to any classifier, not only tree-based classifiers.
    * SFI is not limited to accuracy as the sole performance score.
    * Unlike MDI and MDA, no substitution effects take place, since only one feature is taken into consideration at a time.
    * Like MDA, it can conclude that all features are unimportant, because performance is evaluated via OOS CV.

    The main limitation of SFI is that a classifier with two features can perform better than the bagging of two
    single-feature classifiers. For example, (1) feature B may be useful only in combination with feature A;
    or (2) feature B may be useful in explaining the splits from feature A, even if feature B alone is inaccurate.
    In other words, joint effects and hierarchical importance are lost in SFI. One alternative would be to compute the
    OOS performance score from subsets of features, but that calculation will become intractable as more features are
    considered.

    :param clf: (sklearn.Classifier): Any sklearn classifier.
    :param X: (pd.DataFrame): Train set features.
    :param y: (pd.DataFrame, np.array): Train set labels.
    :param cv_gen: (cross_validation.PurgedKFold): Cross-validation object.
    :param sample_weight_train: (np.array) Sample weights used to train the model for each record in the dataset.
    :param sample_weight_score: (np.array) Sample weights used to evaluate the model quality.
    :param scoring: (function): Scoring function used to determine importance.
    :return: (pd.DataFrame): Mean and standard deviation of feature importance.
    """
    feature_names = X.columns
    if sample_weight_train is None:
        sample_weight_train = np.ones((X.shape[0],))

    if sample_weight_score is None:
        sample_weight_score = np.ones((X.shape[0],))

    imp = pd.DataFrame(columns=['mean', 'std'])
    for feat in feature_names:
        feat_cross_val_scores = ml_cross_val_score(clf, X=X[[feat]], y=y, sample_weight_train=sample_weight_train,
                                                   sample_weight_score=sample_weight_score,
                                                   scoring=scoring, cv_gen=cv_gen)
        imp.loc[feat, 'mean'] = feat_cross_val_scores.mean()
        # pylint: disable=unsubscriptable-object
        imp.loc[feat, 'std'] = feat_cross_val_scores.std() * feat_cross_val_scores.shape[0] ** -.5
    return imp


def plot_feature_importance(importance_df, oob_score, oos_score, save_fig=False, output_path=None):
    """
    Advances in Financial Machine Learning, Snippet 8.10, page 124.

    Feature importance plotting function.

    Plot feature importance.

    :param importance_df: (pd.DataFrame): Mean and standard deviation feature importance.
    :param oob_score: (float): Out-of-bag score.
    :param oos_score: (float): Out-of-sample (or cross-validation) score.
    :param save_fig: (bool): Boolean flag to save figure to a file.
    :param output_path: (str): If save_fig is True, path where figure should be saved.
    """
    # Plot mean imp bars with std
    plt.figure(figsize=(10, importance_df.shape[0] / 5))
    importance_df.sort_values('mean', ascending=True, inplace=True)
    importance_df['mean'].plot(kind='barh', color='b', alpha=0.25, xerr=importance_df['std'], error_kw={'ecolor': 'r'})
    plt.title('Feature importance. OOB Score:{}; OOS score:{}'.format(round(oob_score, 4), round(oos_score, 4)))

    if save_fig is True:
        plt.savefig(output_path)
    else:
        plt.show()



"""
Tests Symmetric Correlation Driven Nonparametric Learning - K.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.scornk \
    import SCORNK


class TestSymmetricCorrelationDrivenNonparametricLearningK(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Symmetric Correlation Driven Nonparametric Learning - K class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_scorn_k_solution(self):
        """
        Test the calculation of SCORN-K.
        """
        # Initialize SCORN-K.
        scorn_k = SCORNK(window=2, rho=2, k=1)
        # Allocates asset prices to SCORN-K.
        scorn_k.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(scorn_k.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_scorn_k_window_error(self):
        """
        Tests ValueError if window is not an integer or less than 1.
        """
        # Initialize SCORN-K.
        scorn_k1 = SCORNK(window=2.5, rho=2, k=1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k1.allocate(self.data)

        # Initialize SCORN-K.
        scorn_k2 = SCORNK(window=0, rho=2, k=1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k2.allocate(self.data)

    def test_scorn_k_rho_error(self):
        """
        Tests ValueError if rho is not an integer or less than 1.
        """
        # Initialize SCORN-K.
        scorn_k3 = SCORNK(window=2, rho=2.5, k=1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k3.allocate(self.data)

        # Initialize SCORN-K.
        scorn_k4 = SCORNK(window=2, rho=0, k=1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k4.allocate(self.data)

    def test_scorn_k_k_error(self):
        """
        Tests ValueError if k is greater than window * rho, greater than 1, or an integer.
        """
        # Initialize SCORN-K.
        scorn_k5 = SCORNK(window=2, rho=2, k=5)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k5.allocate(self.data)

        # Initialize SCORN-K.
        scorn_k6 = SCORNK(window=2, rho=2, k=1.5)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k6.allocate(self.data)

        # Initialize SCORN-K.
        scorn_k7 = SCORNK(window=2, rho=2, k=0)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            scorn_k7.allocate(self.data)



# pylint: disable=missing-module-docstring
import numpy as np
import scipy.optimize as opt
from famlafl.online_portfolio_selection.base import OLPS


class CORN(OLPS):
    """
    This class implements the Correlation Driven Nonparametric Learning strategy. It is
    reproduced with modification from the following paper:
    `Li, B., Hoi, S.C., & Gopalkrishnan, V. (2011). CORN: Correlation-driven nonparametric
    learning approach for portfolio selection. ACM TIST, 2,
    21:1-21:29. <https://dl.acm.org/doi/abs/10.1145/1961189.1961193>`_

    Correlation Driven Nonparametric Learning finds similar windows from the past and looks to
    create portfolio weights that will maximize returns in the similar sets.
    """

    def __init__(self, window, rho):
        """
        Initializes Correlation Driven Nonparametric Learning with the given window and rho value.

        :param window: (int) Number of windows to look back for similarity sets. Windows can be set
                             to any values but typically work well in a shorter term of [1, 7].
        :param rho: (float) Threshold for similarity with range of [-1, 1].
                            Lower rho values will classify more periods as being similar, and higher
                            values will be more strict on identifying a period as similarly correlated.
                            Rho values between [0, 0.2] typically had higher results.
        """
        self.window = window
        self.rho = rho
        self.corr_coef = None  # (np.array) Rolling correlation coefficients.
        super().__init__()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        super(CORN, self)._initialize(asset_prices, weights,
                                      resample_by)
        # Check that window value is an integer.
        if not isinstance(self.window, int):
            raise ValueError("Window value must be an integer.")

        # Check that window value is at least 1.
        if self.window < 1:
            raise ValueError("Window value must be greater than or equal to 1.")

        # Check that rho is between -1 and 1.
        if self.rho < -1 or self.rho > 1:
            raise ValueError("Rho must be between -1 and 1.")
        self.corr_coef = self._calc_rolling_corr_coef()

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # Create similar set.
        similar_set = []

        # Default is uniform weights.
        new_weights = self._uniform_weight()

        # Calculate for similar sets if time is greater or equal to window size.
        if time >= self.window:
            # Iterate through past windows.
            for past_time in range(time - self.window + 1):
                # Check for windows that are above rho.
                if self.corr_coef[time - self.window + 1][past_time] > self.rho:
                    # Append the time for similar set.
                    similar_set.append(past_time + self.window)

            if similar_set:
                # Choose the corresponding relative return periods.
                optimize_array = self.relative_return[similar_set]
                new_weights = self._fast_optimize(optimize_array)
        return new_weights

    def _fast_optimize(self, optimize_array):
        """
        Calculates weights that maximize returns over the given array.

        :param optimize_array: (np.array) Relative returns of the assets for a given time period.
        :return: (np.array) Weights that maximize the returns for the given array.
        """
        # Initialize guess.
        weights = self._uniform_weight()

        # Use np.log and np.sum to make the cost function a convex function.
        # Multiplying continuous returns equates to summing over the log returns.
        def _objective(weight):
            return -np.sum(np.log(np.dot(optimize_array, weight)))

        # Derivative of the objective function.
        def _derivative(weight):
            total_returns = np.dot(optimize_array, weight)
            return -np.dot(1 / total_returns, optimize_array)

        # Weight bounds.
        bounds = tuple((0.0, 1.0) for asset in range(self.number_of_assets))

        # Sum of weights is 1.
        const = ({'type': 'eq', 'fun': lambda w: np.sum(w) - 1})

        problem = opt.minimize(_objective, weights, method='SLSQP', bounds=bounds,
                               constraints=const, jac=_derivative)
        return problem.x

    def _calc_rolling_corr_coef(self):
        """
        Calculates the rolling correlation coefficient for a given relative return and window

        :return: (np.array) Rolling correlation coefficient over a given window.
        """
        # Flatten the array.
        flattened = self.relative_return.flatten()

        # Set index of rolled window.
        idx = np.arange(self.number_of_assets * self.window)[None, :] + self.number_of_assets * \
              np.arange(self.length_of_time - self.window + 1)[:, None]

        # Retrieve the results of the rolled window.
        rolled_returns = flattened[idx]

        # Calculate correlation coefficient.
        with np.errstate(divide='ignore', invalid='ignore'):
            rolling_corr_coef = np.nan_to_num(np.corrcoef(rolled_returns), nan=0)
        return rolling_corr_coef



# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd
import scipy.spatial.distance as ssd
import scipy.cluster.hierarchy as sch
from famlafl.portfolio_optimization.risk_estimators import RiskEstimators


class TIC:
    """
    This class implements the Theory-Implied Correlation (TIC) algorithm and the correlation matrix distance
    introduced by Herdin and Bonek. It is reproduced with modification from the following paper:
    `Marcos Lopez de Prado “Estimation of Theory-Implied Correlation Matrices”, (2019).
    <https://papers.ssrn.com/abstract_id=3484152>`_.
    """

    def __init__(self):
        """
        Initialize
        """

        return

    def tic_correlation(self, tree_struct, corr_matrix, tn_relation, kde_bwidth=0.01):
        """
        Calculates the Theory-Implied Correlation (TIC) matrix.

        Includes three steps.

        In the first step, the theoretical tree graph structure of the assets is fit on the evidence
        presented by the empirical correlation matrix.

        The result of the first step is a binary tree (dendrogram) that sequentially clusters two elements
        together, while measuring how closely together the two elements are, until all elements are
        subsumed within the same cluster.

        In the second step, a correlation matrix is derived from the linkage object.

        Each cluster in the global linkage object is decomposed into two elements,
        which can be either atoms or other clusters. Then the off-diagonal correlation between two
        elements is calculated based on the distances between them.

        In the third step, the correlation matrix is de-noised.

        This is done by fitting the Marcenko-Pastur distribution to the eigenvalues of the matrix, calculating the
        maximum theoretical eigenvalue as a threshold and shrinking the eigenvalues higher than a set threshold.
        This algorithm is implemented in the RiskEstimators class.

        :param tree_struct: (pd.dataframe) The tree graph that represents the structure of the assets
        :param corr_matrix: (pd.dataframe) The empirical correlation matrix of the assets
        :param tn_relation: (float) Relation of sample length T to the number of variables N used to calculate the
                                    correlation matrix
        :param kde_bwidth: (float) The bandwidth of the kernel to fit KDE for de-noising the correlation matrix
                                   (0.01 by default)
        :return: (np.array) Theory-Implied Correlation matrix
        """

        # Getting the linkage object that characterizes the dendrogram
        lnkage_object = self._get_linkage_corr(tree_struct, corr_matrix)

        # Calculating the correlation matrix from the dendrogram
        ti_correlation = self._link2corr(lnkage_object, corr_matrix.index)

        # Class with function for de-noising the correlation matrix
        risk_estim = RiskEstimators()

        # De-noising the obtained Theory-Implies Correlation matrix
        ti_correlation_denoised = risk_estim.denoise_covariance(ti_correlation, tn_relation=tn_relation,
                                                                kde_bwidth=kde_bwidth)

        return ti_correlation_denoised

    @staticmethod
    def corr_dist(corr0, corr1):
        """
        Calculates the correlation matrix distance proposed by Herdin and Bonek.

        The distance obtained measures the orthogonality between the considered
        correlation matrices. If the matrices are equal up to a scaling factor,
        the distance becomes zero and one if they are different to a maximum
        extent.

        This can be used to measure to which extent the TIC matrix has blended
        theory-implied views (tree structure of the elements) with empirical
        evidence (correlation matrix).

        :param corr0: (pd.dataframe) First correlation matrix
        :param corr1: (pd.dataframe) Second correlation matrix
        :return: (float) Correlation matrix distance
        """

        # Trace of the product of correlation matrices
        prod_trace = np.trace(np.dot(corr0, corr1))

        # Frobenius norm of the first correlation matrix
        frob_product = np.linalg.norm(corr0, ord='fro')

        # Frobenius norm of the second correlation matrix
        frob_product *= np.linalg.norm(corr1, ord='fro')

        # Distance calculation
        distance = 1 - prod_trace / frob_product

        return distance

    def _get_linkage_corr(self, tree_struct, corr_matrix):
        """
        Fits the theoretical tree graph structure of the assets in a portfolio on the evidence
        presented by the empirical correlation matrix.

        The result is a binary tree (dendrogram) that sequentially clusters two elements
        together, while measuring how closely together the two elements are, until all elements are
        subsumed within the same cluster.

        This is the first step of the TIC algorithm.

        :param tree_struct: (pd.dataframe) The tree graph that represents the structure of the assets
        :param corr_matrix: (pd.dataframe) The empirical correlation matrix of the assets
        :return: (np.array) Linkage object that characterizes the dendrogram
        """

        # If the top level of the tree contains multiple elements, creating a level with just one element (tree root)
        if len(np.unique(tree_struct.iloc[:, -1])) > 1:
            tree_struct = tree_struct.copy(deep=True)
            tree_struct['All'] = 0

        # Creating a linkage object (matrix with link elements).
        # Each row represents a cluster and consists of the following columns:
        # (1) and (2) - number ids of two elements clustered together
        # (3) - the distance between those elements
        # (4) - number of original variables in this cluster
        # Items in a cluster can be single elements or other clusters
        global_linkage = np.empty(shape=(0, 4))

        # List with elements containing two consecutive tree levels
        tree_levels = [[tree_struct.columns[i-1], tree_struct.columns[i]] for i in range(1, tree_struct.shape[1])]

        # Calculating the distance matrix from the empirical correlation matrix
        distance_matrix = ((1 - corr_matrix) / 2)**(1/2)

        # Getting a list of names of elements
        global_elements = distance_matrix.index.tolist()

        # Iterating through levels of the tree
        for level in tree_levels:

            # Taking two consecutive levels of the tree
            # Removing duplicates from the lower level of the tree
            # Setting the obtained unique elements from the lower level as an index
            # Grouping by elements in the higher level
            grouped_level = tree_struct[level].drop_duplicates(level[0]).set_index(level[0]).groupby(level[1])

            # Iterating through the obtained two levels of a tree
            for high_element, grouped_elements in grouped_level:
                # high_element contains the higher element
                # grouped_elements contain the elements from the lower level, grouped under the higher-level element

                # Getting the names of the grouped elements
                grouped_elements = grouped_elements.index.tolist()

                # If only one element grouped by the element from the higher level
                if len(grouped_elements) == 1:
                    # Changing the name of this element to the name of the element from the higher level
                    # As this element is now representing the higher level

                    # Changing the name in the list of names
                    global_elements[global_elements.index(grouped_elements[0])] = high_element

                    # Changing the name also in the correlation matrix rows and columns
                    distance_matrix = distance_matrix.rename({grouped_elements[0]: high_element}, axis=0)
                    distance_matrix = distance_matrix.rename({grouped_elements[0]: high_element}, axis=1)

                    continue

                # Taking the part of the distance matrix containing the grouped elements in the lower level
                local_distance = distance_matrix.loc[grouped_elements, grouped_elements]

                # Transforming distance matrix to distance vector
                # Check for matrix symmetry is made - checking that the matrix given is a distance matrix
                distance_vec = ssd.squareform(local_distance, force='tovector',
                                              checks=(not np.allclose(local_distance, local_distance.T)))

                # Doing hierarchical clustering of the distance vector. Result is a linkage object
                # Here we have created new clusters based on the grouped elements
                local_linkage = sch.linkage(distance_vec, optimal_ordering=True)

                # Transforming the linkage object from local (only containing the grouped elements)
                # to global (containing all elements) form
                local_linkage_transformed = self._link_clusters(global_linkage, local_linkage, global_elements,
                                                                grouped_elements)

                # Adding new link elements to the general linage object
                global_linkage = np.append(global_linkage, local_linkage_transformed, axis=0)

                # As more clusters were created, their names are added to the global list of elements
                global_elements += range(len(global_elements), len(global_elements) + len(local_linkage_transformed))

                # Updating the general distance matrix to take the new clusters into account
                # Now the grouped elements in the distance matrix will be replaced with new clusters as elements.
                distance_matrix = self._update_dist(distance_matrix, global_linkage, local_linkage_transformed,
                                                    global_elements)

                # The last added cluster is representing the element from the higher level
                # So we're changing the name of that cluster to the name of the higher level element
                global_elements[-1] = high_element

                # Changing the name of the cluster also in the distance matrix
                distance_matrix.columns = distance_matrix.columns[:-1].tolist() + [high_element]
                distance_matrix.index = distance_matrix.columns

        # Changing the linkage object from an array of arrays to array of tuples with named fields
        global_linkage = np.array([*map(tuple, global_linkage)],
                                  dtype=[('i0', int), ('i1', int), ('dist', float), ('num', int)])

        return global_linkage

    @staticmethod
    def _link_clusters(global_linkage, local_linkage, global_elements, grouped_elements):
        """
        Transforms linkage object from local local_linkage (based on dist1) into global global_linkage (based on dist0)

        Consists of changes of names for the elements in clusters and change of the number of
        basic elements (atoms) contained inside a cluster. This is done to take into account the
        already existing links.

        :param global_linkage: (np.array) Global linkage object (previous links)
        :param local_linkage: (np.array) Local linkage object (containing grouped elements and not global ones)
        :param global_elements: (list) List of names for all elements (global)
        :param grouped_elements: (list) List of grouped elements (local)
        :return: (np.array) Local linkage object changed to global one
        """

        # Counting the number of atoms - basic elements and not clusters
        # It's the total number of elements minus the number of links (each link represents 1 cluster)
        num_atoms = len(global_elements) - global_linkage.shape[0]

        # Making a copy of a local linkage object
        local_linkage_tr = local_linkage.copy()

        # Iterating through links in the local linkage object
        for link in range(local_linkage_tr.shape[0]):
            # Counting the number of atoms in the cluster (represented by this link)
            atom_counter = 0

            # Iterating through the two elements contained in a cluster (represented by this link)
            for j in range(2):
                # Changing the names in links to global ones

                if local_linkage_tr[link, j] < len(grouped_elements):  # If it's the element from the grouped ones
                    # Then replacing its local name with the actual name from the list of all elements' names
                    local_linkage_tr[link, j] = global_elements.index(grouped_elements[int(local_linkage_tr[link, j])])

                else:  # Otherwise it's a new cluster
                    # Then giving it a new name, taking into account the previously named clusters
                    # The names of the clusters are sequential numbers
                    local_linkage_tr[link, j] += -len(grouped_elements) + len(global_elements)

                # Updating the number of atoms in a cluster (represented by this link)

                if local_linkage_tr[link, j] < num_atoms:  # If the added element is an atom
                    # Then add one to the counter of atoms inside
                    atom_counter += 1

                else:  # If the element added is a cluster
                    # If the added element is a previously created cluster
                    if local_linkage_tr[link, j] - num_atoms < global_linkage.shape[0]:
                        # Adding to counter the number of atoms from the global linkage object
                        atom_counter += global_linkage[int(local_linkage_tr[link, j]) - num_atoms, 3]

                    else:  # If the added element is a newly created cluster
                        # Adding to the counter the number of atoms from the local linkage object
                        atom_counter += local_linkage_tr[int(local_linkage_tr[link, j]) - len(global_elements), 3]

            # Setting the number of atoms in the cluster to the calculated counter
            local_linkage_tr[link, 3] = atom_counter

        return local_linkage_tr

    @staticmethod
    def _update_dist(distance_matrix, global_linkage, local_linkage_tr, global_elements, criterion=None):
        """
        Updates the general distance matrix to take the new clusters into account

        Replaces the elements added to the new clusters with these clusters as elements.
        Requires the recalculation of the distance matrix to determine the distance from
        new clusters to other elements.

        A criterion function may be given for calculation of the new distances from a new cluster to other
        elements based on the distances of elements included in a cluster. The default method is the weighted
        average of distances based on the number of atoms in each of the two elements.

        :param distance_matrix: (pd.dataframe) Previous distance matrix
        :param global_linkage: (np.array) Global linkage object that includes new clusters
        :param local_linkage_tr: (np.array) Local linkage object transformed (global names of elements and atoms count)
        :param global_elements: (list) Global list with names of all elements
        :param criterion: (function) Function to apply to a dataframe of distances to adjust them
        :return: (np.array) Updated distance matrix
        """

        # Counting the number of atoms - basic elements  and not clusters
        num_atoms = len(global_elements) - global_linkage.shape[0]

        # Getting the list with names of new items
        new_items = global_elements[-local_linkage_tr.shape[0]:]

        # Iterating through elements in the local linkage object
        for i in range(local_linkage_tr.shape[0]):
            # Getting the names of two elements clustered together
            elem_1, elem_2 = global_elements[int(local_linkage_tr[i, 0])], global_elements[int(local_linkage_tr[i, 1])]

            # If no criterion function given to determine new distances then the weighted average
            # based on the number of atoms in each of the two elements is used
            if criterion is None:

                if local_linkage_tr[i, 0] < num_atoms:  # If the first element is an atom
                    # Weight of the element is 1
                    elem_1_weight = 1

                else:  # If the first element is a cluster
                    # Weight is set to the number of atoms in a cluster
                    elem_1_weight = global_linkage[int(local_linkage_tr[i, 0]) - num_atoms, 3]

                if local_linkage_tr[i, 1] < num_atoms:  # If the second element is an atom
                    # Weight of the element is 1
                    elem_2_weight = 1

                else:  # If the second element is a cluster
                    # Weight is set to the number of atoms in a cluster
                    elem_2_weight = global_linkage[int(local_linkage_tr[i, 1]) - num_atoms, 3]

                # Calculating new distance as the average weighted distance
                # where the weight is the number of atoms in an element
                dist_vector = (distance_matrix[elem_1] * elem_1_weight + distance_matrix[elem_2] * elem_2_weight) / \
                              (elem_1_weight + elem_2_weight)

            # If a criterion function is given, the new distance is calculated using it
            else:
                # New distance
                dist_vector = criterion(distance_matrix[[elem_1, elem_2]], axis=1)

            # Adding a column with th new cluster to the distance matrix
            distance_matrix[new_items[i]] = dist_vector

            # Adding row with the new cluster to the distance matrix
            distance_matrix.loc[new_items[i]] = dist_vector

            # Setting the main diagonal value for the new cluster to 0 (the distance of the element to itself is zero)
            distance_matrix.loc[new_items[i], new_items[i]] = 0

            # And deleting the two elements that were combined in the new cluster
            distance_matrix = distance_matrix.drop([elem_1, elem_2], axis=0)
            distance_matrix = distance_matrix.drop([elem_1, elem_2], axis=1)

        return distance_matrix

    @staticmethod
    def _get_atoms(linkage, element):
        """
        Getting the atoms included in an element from a linkage object

        Atoms are the basic assets in a portfolio and not clusters.

        :param linkage: (np.array) Global linkage object
        :param element: (int) Element id to get atoms from
        :return: (list) Set of atoms
        """

        # A list of elements to unpack
        # Now includes only one given element but will be appended with the unpacked elements
        element_list = [element]

        # Iterating (until there are elements to unpack)
        while True:
            # The maximum item from the list (as the clusters have higher numbers in comparison to atoms)
            item_ = max(element_list)

            # If it's a cluster and not an atom
            if item_ > linkage.shape[0]:
                # Delete this cluster
                element_list.remove(item_)

                # Unpack the elements in a cluster and add them to a list of elements to unpack
                element_list.append(linkage['i0'][item_ - linkage.shape[0] - 1])
                element_list.append(linkage['i1'][item_ - linkage.shape[0] - 1])

            else:  # If all the elements left in the list are atoms, we're done
                break

        # The resulting list contains only atoms
        return element_list

    def _link2corr(self, linkage, element_index):
        """
        Derives a correlation matrix from the linkage object.

        Each cluster in the global linkage object is decomposed into two elements,
        which can be either atoms or other clusters. Then the off-diagonal correlation between two
        elements are calculated based on the distances between them.

        This is the second step of the TIC algorithm.

        :param linkage: (np.array) Global linkage object
        :param element_index: (pd.index) Names of elements used to calculate the linkage object
        :return: (pd.dataframe) Correlation matrix associated with linkage object
        """

        # Creating a base for new correlation matrix with ones on the main diagonal
        corr_matrix = pd.DataFrame(np.eye(linkage.shape[0]+1), index=element_index, columns=element_index, dtype=float)

        # Iterating through links in the linkage object
        for link in range(linkage.shape[0]):
            # Getting the atoms contained in the first element from the link
            el_x = self._get_atoms(linkage, linkage['i0'][link])

            # Getting the atoms contained in the second element from the link
            el_y = self._get_atoms(linkage, linkage['i1'][link])

            # Calculating the odd-diagonal values of the correlation matrix
            corr_matrix.loc[element_index[el_x], element_index[el_y]] = 1 - 2 * linkage['dist'][link]**2

            # And the symmetrical values
            corr_matrix.loc[element_index[el_y], element_index[el_x]] = 1 - 2 * linkage['dist'][link]**2

        return corr_matrix



# pylint: disable=missing-module-docstring
import numpy as np
import scipy.optimize as opt
from famlafl.online_portfolio_selection.corn import CORN


class FCORN(CORN):
    """
    This class implements the Functional Correlation Driven Nonparametric Learning strategy. It is
    reproduced with modification from the following paper:
    `Wang, Y., & Wang, D. (2019). Market Symmetry and Its Application to Pattern-Matching-Based
    Portfolio Selection. The Journal of Financial Data Science, 1(2), 78–92.
    <https://jfds.pm-research.com/content/1/2/78>`_

    Functional Correlation Driven Nonparametric Learning is an extension of the SCORN strategy.
    Unlike SCORN that sets a certain threshold for the objective function, FCORN places an
    arbitrary weight to all relative returns period to emphasize certain weights. If the given
    lambda value approaches infinity, then FCORN returns the same values as the SCORN strategy.
    """

    def __init__(self, window, rho, lambd):
        """
        Initializes Functional Correlation Driven Nonparametric Learning with the given window,
        rho, and lambd value.

        :param window: (int) Number of windows to look back for similarity sets. Windows can be set
                             to any values but typically work well in a shorter term of [1, 7].
        :param rho: (float) Threshold for similarity. Rho should set in the range of [-1, 1].
                            Lower rho values will classify more periods as being similar, and higher
                            values will be more strict on identifying a period as similarly correlated.
                            Rho values between [0.4, 0.8] typically had higher results.
        :param lambd: (float) Scale factor for the sigmoid activation function. Lambd can be any
                              value. If lambd approaches infinity, FCORN will have the same results
                              as SCORN, and if lambd approaches negative infinity, FCORN produces
                              similar results as a Constant Rebalanced Portfolio. From experimental
                              results, lambd of 1 was optimal for some datasets.
        """
        self.lambd = lambd
        super().__init__(window=window, rho=rho)

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # Default is uniform weights.
        new_weights = self._uniform_weight()

        # Calculate for similar sets if time is greater or equal to window size.
        if time >= self.window:
            # Create activation_fn to scale the relative returns during the optimization process.
            activation_fn = np.zeros(time + 1)

            # Iterate through past windows.
            for past_time in range(time - self.window + 1):
                # Get correlation between the two windows.
                corr = self.corr_coef[time - self.window + 1][past_time]

                # If correlation is non-negative.
                if corr >= 0:
                    # Set corresponding activation function
                    activation_fn[past_time + self.window] = self._sigmoid(
                        self.lambd * (corr - self.rho))

                # If correlation is negative.
                else:
                    activation_fn[past_time + self.window] = self._sigmoid(
                        self.lambd * (corr + self.rho)) - 1

            # Negative activation function to minimize final objective function.
            activation_fn = -activation_fn

            # Current relative returns.
            curr_time = np.asfortranarray(self.relative_return[:time + 1])

            # Calculate new weights based on activation_fn and relatives returns to date.
            new_weights = self._fcorn_optimize(activation_fn, curr_time)
        return new_weights

    def _fcorn_optimize(self, activation_fn, relative_return):
        """
        Calculates weights that maximize returns over the given array.

        :param activation_fn: (np.array) Scaling factor for optimization equation.
        :param relative_return: (np.array) Array of relative returns until the current period
        :return: (np.array) Weights that maximize the returns for the given array.
        """
        # Initialize guess.
        weights = self._uniform_weight()

        # Use np.log and np.sum to make the cost function a convex function.
        # Multiplying continuous returns equates to summing over the log returns.
        def _objective(weight):
            return np.dot(activation_fn, np.log(np.dot(relative_return, weight)))

        def _derivative(weight):
            total_returns = np.dot(relative_return, weight)
            return np.dot(activation_fn / total_returns, relative_return)

        # Weight bounds.
        bounds = tuple((0.0, 1.0) for asset in range(self.number_of_assets))

        # Sum of weights is 1.
        const = ({'type': 'eq', 'fun': lambda w: np.sum(w) - 1.0})

        problem = opt.minimize(_objective, weights, method='SLSQP', bounds=bounds,
                               constraints=const, jac=_derivative)
        return problem.x

    @staticmethod
    def _sigmoid(val):
        """
        Generates the resulting sigmoid function

        :param val: (float) Input for
        :return: (float) sigmoid(x)
        """
        # Calculate sigmoid.
        res = 1 / (1 + np.exp(-val))
        return res



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.cornk import CORNK
from famlafl.online_portfolio_selection.scorn import SCORN


class SCORNK(CORNK):
    """
    This class implements the Symmetric Correlation Driven Nonparametric Learning - K strategy. It
    is reproduced with modification from the following paper:
    `Wang, Y., & Wang, D. (2019). Market Symmetry and Its Application to Pattern-Matching-Based
    Portfolio Selection. The Journal of Financial Data Science, 1(2), 78–92.
    <https://jfds.pm-research.com/content/1/2/78>`_

    Symmetric Correlation Driven Nonparametric Learning - K formulates a number of SCORN experts and
    tracks the experts performance over time. Each period, the strategy decides to allocate capital
    to the top-k experts until the previous time period. This strategy takes an ensemble approach to
    the top performing experts.
    """

    def _generate_experts(self):
        """
        Generates window * rho experts from window of [1, w] and rho of [0, (rho - 1) / rho].
        """
        # Initialize expert parameters.
        self.expert_params = np.zeros((self.number_of_experts, 2))

        # Pointer to iterate through parameter locations.
        pointer = 0

        # Window from 1 to self.window.
        for n_window in range(self.window):
            # Rho from 0 to (rho - 1)/rho.
            for n_rho in range(self.rho):
                # Assign experts with parameters (n_window + 1, n_rho/rho).
                self.expert_params[pointer] = [n_window + 1, n_rho/self.rho]
                # Next pointer.
                pointer += 1

        for exp in range(self.number_of_experts):
            param = self.expert_params[exp]
            self.experts.append(SCORN(int(param[0]), param[1]))



"""
Various useful functions
"""

import pandas as pd
import numpy as np

def crop_data_frame_in_batches(df: pd.DataFrame, chunksize: int):
    # pylint: disable=invalid-name
    """
    Splits df into chunks of chunksize

    :param df: (pd.DataFrame) Dataframe to split
    :param chunksize: (int) Number of rows in chunk
    :return: (list) Chunks (pd.DataFrames)
    """
    generator_object = []
    for _, chunk in df.groupby(np.arange(len(df)) // chunksize):
        generator_object.append(chunk)
    return generator_object



# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd
from sklearn.covariance import LedoitWolf
from sklearn.cluster import KMeans
from sklearn.metrics import silhouette_samples
from scipy.linalg import block_diag
from famlafl.portfolio_optimization.risk_estimators import RiskEstimators


class NCO:
    """
    This class implements the Nested Clustered Optimization (NCO) algorithm, the Convex Optimization Solution (CVO),
    the  Monte Carlo Optimization Selection (MCOS) algorithm and sample data generating function. It is reproduced with
    modification from the following paper: `Marcos Lopez de Prado “A Robust Estimator of the Efficient Frontier”,
    (2019). <https://papers.ssrn.com/abstract_id=3469961>`_.
    """

    def __init__(self):
        """
        Initialize
        """

        return

    @staticmethod
    def allocate_cvo(cov, mu_vec=None):
        """
        Estimates the Convex Optimization Solution (CVO).

        Uses the covariance matrix and the mu - optimal solution.
        If mu is the vector of expected values from variables, the result will be
        a vector of weights with maximum Sharpe ratio.
        If mu is a vector of ones, the result will be a vector of weights with
        minimum variance.

        :param cov: (np.array) Covariance matrix of the variables.
        :param mu_vec: (np.array) Expected value of draws from the variables for maximum Sharpe ratio.
                              None if outputting the minimum variance portfolio.
        :return: (np.array) Weights for optimal allocation.
        """

        # Calculating the inverse covariance matrix
        inv_cov = np.linalg.inv(cov)

        # Generating a vector of size of the inverted covariance matrix
        ones = np.ones(shape=(inv_cov.shape[0], 1))

        if mu_vec is None:  # To output the minimum variance portfolio
            mu_vec = ones

        # Calculating the analytical solution using CVO - weights
        w_cvo = np.dot(inv_cov, mu_vec)
        w_cvo /= np.dot(mu_vec.T, w_cvo)

        return w_cvo

    def allocate_nco(self, cov, mu_vec=None, max_num_clusters=None, n_init=10):
        """
        Estimates the optimal allocation using the nested clustered optimization (NCO) algorithm.

        First, it clusters the covariance matrix into subsets of highly correlated variables.
        Second, it computes the optimal allocation for each of the clusters separately.
        This allows collapsing of the original covariance matrix into a reduced covariance matrix,
        where each cluster is represented by a single variable.
        Third, we compute the optimal allocations across the reduced covariance matrix.
        Fourth, the final allocations are the dot-product of the intra-cluster (step 2) allocations and
        the inter-cluster (step 3) allocations.

        For the Convex Optimization Solution (CVO), a mu - optimal solution parameter is needed.
        If mu is the vector of expected values from variables, the result will be
        a vector of weights with maximum Sharpe ratio.
        If mu is a vector of ones (pass None value), the result will be a vector of weights with
        minimum variance.

        :param cov: (np.array) Covariance matrix of the variables.
        :param mu_vec: (np.array) Expected value of draws from the variables for maximum Sharpe ratio.
                              None if outputting the minimum variance portfolio.
        :param max_тum_сlusters: (int) Allowed maximum number of clusters. If None then taken as num_elements/2.
        :param n_init: (float) Number of time the k-means algorithm will run with different centroid seeds (default 10)
        :return: (np.array) Optimal allocation using the NCO algorithm.
        """

        # Using pd.DataFrame instead of np.array
        cov = pd.DataFrame(cov)

        # Optimal solution for minimum variance
        if mu_vec is not None:
            mu_vec = pd.Series(mu_vec[:, 0])

        # Class with function to calculate correlation from covariance function
        risk_estimators = RiskEstimators()

        # Calculating correlation matrix
        corr = risk_estimators.cov_to_corr(cov)

        # Optimal partition of clusters (step 1)
        corr, clusters, _ = self._cluster_kmeans_base(corr, max_num_clusters, n_init=n_init)

        # Weights inside clusters
        w_intra_clusters = pd.DataFrame(0, index=cov.index, columns=clusters.keys())

        # Iterating over clusters
        for i in clusters:
            # Covariance matrix of elements in cluster
            cov_cluster = cov.loc[clusters[i], clusters[i]].values

            # Optimal solution vector for the cluster
            mu_cluster = (None if mu_vec is None else mu_vec.loc[clusters[i]].values.reshape(-1, 1))

            # Estimating the Convex Optimization Solution in a cluster (step 2)
            w_intra_clusters.loc[clusters[i], i] = self.allocate_cvo(cov_cluster, mu_cluster).flatten()

        # Reducing new covariance matrix to calculate inter-cluster weights
        cov_inter_cluster = w_intra_clusters.T.dot(np.dot(cov, w_intra_clusters))
        mu_inter_cluster = (None if mu_vec is None else w_intra_clusters.T.dot(mu_vec))

        # Optimal allocations across the reduced covariance matrix (step 3)
        w_inter_clusters = pd.Series(self.allocate_cvo(cov_inter_cluster, mu_inter_cluster).flatten(),
                                     index=cov_inter_cluster.index)

        # Final allocations - dot-product of the intra-cluster and inter-cluster allocations (step 4)
        w_nco = w_intra_clusters.mul(w_inter_clusters, axis=1).sum(axis=1).values.reshape(-1, 1)

        return w_nco

    def allocate_mcos(self, mu_vec, cov, num_obs, num_sims=100, kde_bwidth=0.01, min_var_portf=True, lw_shrinkage=False):
        """
        Estimates the optimal allocation using the Monte Carlo optimization selection (MCOS) algorithm.

        Repeats the CVO and the NCO algorithms multiple times on the empirical values to get a dataframe of trials
        in order to later compare them to a true optimal weights allocation and compare the robustness of the NCO
        and CVO methods.

        :param mu_vec: (np.array) The original vector of expected outcomes.
        :param cov: (np.array )The original covariance matrix of outcomes.
        :param num_obs: (int) The number of observations T used to compute mu_vec and cov.
        :param num_sims: (int) The number of Monte Carlo simulations to run. (100 by default)
        :param kde_bwidth: (float) The bandwidth of the KDE used to de-noise the covariance matrix. (0.01 by default)
        :param min_var_portf: (bool) When True, the minimum variance solution is computed. Otherwise, the
                                     maximum Sharpe ratio solution is computed. (True by default)
        :param lw_shrinkage: (bool) When True, the covariance matrix is subjected to the Ledoit-Wolf shrinkage
                                    procedure. (False by default)
        :return: (pd.DataFrame, pd.DataFrame) DataFrames with allocations for CVO and NCO algorithms.
        """

        # Creating DataFrames for CVO results and NCO results
        w_cvo = pd.DataFrame(columns=range(cov.shape[0]), index=range(num_sims), dtype=float)
        w_nco = w_cvo.copy(deep=True)

        # Iterating thorough simulations
        for simulation in range(num_sims):
            # Deriving empirical vector of means and an empirical covariance matrix
            mu_simulation, cov_simulation = self._simulate_covariance(mu_vec, cov, num_obs, lw_shrinkage)

            # If goal is minimum variance
            if min_var_portf:
                mu_simulation = None

            # Class with de-noising function
            risk_estimators = RiskEstimators()

            # De-noising covariance matrix
            if kde_bwidth > 0:
                cov_simulation = risk_estimators.denoise_covariance(cov_simulation, num_obs / cov_simulation.shape[1],
                                                                    kde_bwidth)

            # Writing the results to corresponding dataframes
            w_cvo.loc[simulation] = self.allocate_cvo(cov_simulation, mu_simulation).flatten()
            w_nco.loc[simulation] = self.allocate_nco(cov_simulation, mu_simulation,
                                                      int(cov_simulation.shape[0] / 2)).flatten()

        return w_cvo, w_nco

    def estim_errors_mcos(self, w_cvo, w_nco, mu_vec, cov, min_var_portf=True):
        """
        Computes the true optimal allocation w, and compares that result with the estimated ones by MCOS.

        The result is the mean standard deviation between the true weights and the ones obtained from the simulation
        for each algorithm - CVO and NCO.

        :param w_cvo: (pd.DataFrame) DataFrame with weights from the CVO algorithm.
        :param w_nco: (pd.DataFrame) DataFrame with weights from the NCO algorithm.
        :param mu_vec: (np.array) The original vector of expected outcomes.
        :param cov: (np.array)The original covariance matrix of outcomes.
        :param min_var_portf: (bool) When True, the minimum variance solution was computed. Otherwise, the
                                     maximum Sharpe ratio solution was computed. (True by default)
        :return: (float, float) Mean standard deviation of weights for CVO and NCO algorithms.
        """

        # Calculating the true optimal weights allocation
        w_true = self.allocate_cvo(cov, None if min_var_portf else mu_vec)
        w_true = np.repeat(w_true.T, w_cvo.shape[0], axis=0)

        # Mean standard deviation between the weights from CVO algorithm and the true weights
        err_cvo = (w_cvo - w_true).std(axis=0).mean()

        # Mean standard deviation between the weights from NCO algorithm and the true weights
        err_nco = (w_nco - w_true).std(axis=0).mean()

        return err_cvo, err_nco

    @staticmethod
    def _simulate_covariance(mu_vector, cov_matrix, num_obs, lw_shrinkage=False):
        """
        Derives an empirical vector of means and an empirical covariance matrix.

        Based on the set of true means vector and covariance matrix of X distributions,
        the function generates num_obs observations for every X.
        Based on these observations simulated vector of means and the simulated covariance
        matrix are obtained.

        :param mu_vector: (np.array) True means vector for X distributions
        :param cov_matrix: (np.array) True covariance matrix for X distributions
        :param num_obs: (int) Number of observations to draw for every X
        :param lw_shrinkage: (bool) Flag to apply Ledoit-Wolf shrinkage to X (False by default)
        :return: (np.array, np.array) Empirical means vector, empirical covariance matrix
        """

        # Generating a matrix of num_obs observations for X distributions
        observations = np.random.multivariate_normal(mu_vector.flatten(), cov_matrix, size=num_obs)

        # Empirical means vector calculation
        mu_simulated = observations.mean(axis=0).reshape(-1, 1)

        if lw_shrinkage:  # If applying Ledoit-Wolf shrinkage
            cov_simulated = LedoitWolf().fit(observations).covariance_

        else:  # Simple empirical covariance matrix
            cov_simulated = np.cov(observations, rowvar=False)

        return mu_simulated, cov_simulated

    @staticmethod
    def _cluster_kmeans_base(corr, max_num_clusters=None, n_init=10):
        """
        Finding the optimal partition of clusters using K-Means algorithm.

        For the fit of K-Means algorithm a matrix of distances based on the correlation matrix is used.
        The algorithm iterates n_init number of times and initialises K-Means with different seeds
        and max_number_of_clusters.

        The Silhouette Coefficient is used as a measure of how well samples are clustered
        with samples that are similar to themselves.

        :param corr: (pd.DataFrame) DataFrame with correlation matrix
        :param max_num_clusters: (float) Maximum allowed number of clusters. If None then taken as num_elements/2
        :param n_init: (float) Number of time the k-means algorithm will run with different centroid seeds (default 10)
        :return: (np.array, dict, pd.Series) Correlation matrix of clustered elements, dict with clusters,
                                             Silhouette Coefficient series
        """

        # Distance matrix from correlation matrix
        dist_matrix = ((1 - corr.fillna(0)) / 2) ** (1 / 2)

        # Series for Silhouette Coefficients - cluster fit measure
        silh_coef_optimal = pd.Series(dtype='float64')

        # If maximum number of clusters undefined, it's equal to half the number of elements
        if max_num_clusters is None:
            max_num_clusters = round(corr.shape[0] / 2)

        # Iterating over the allowed iteration times for k-means
        for init in range(1, n_init):
            # Iterating through every number of clusters
            for num_clusters in range(2, max_num_clusters + 1):
                # Computing k-means clustering
                kmeans = KMeans(n_clusters=num_clusters, n_init=init)
                kmeans = kmeans.fit(dist_matrix)

                # Computing a Silhouette Coefficient - cluster fit measure
                silh_coef = silhouette_samples(dist_matrix, kmeans.labels_)

                # Metrics to compare numbers of clusters
                stat = (silh_coef.mean() / silh_coef.std(), silh_coef_optimal.mean() / silh_coef_optimal.std())

                # If this is the first metric or better than the previous
                # we set it as the optimal number of clusters
                if np.isnan(stat[1]) or stat[0] > stat[1]:
                    silh_coef_optimal = silh_coef
                    kmeans_optimal = kmeans

        # Sorting labels of clusters
        new_index = np.argsort(kmeans_optimal.labels_)

        # Reordering correlation matrix rows
        corr = corr.iloc[new_index]

        # Reordering correlation matrix columns
        corr = corr.iloc[:, new_index]

        # Preparing cluster members as dict
        clusters = {i: corr.columns[np.where(kmeans_optimal.labels_ == i)[0]].tolist() for \
                    i in np.unique(kmeans_optimal.labels_)}

        # Silhouette Coefficient series
        silh_coef_optimal = pd.Series(silh_coef_optimal, index=dist_matrix.index)

        return corr, clusters, silh_coef_optimal

    @staticmethod
    def _form_block_matrix(num_blocks, block_size, block_corr):
        """
        Creates a correlation matrix in a block form with given parameters.

        :param num_blocks: (int) Number of blocks in matrix
        :param block_size: (int) Size of a single block
        :param block_corr: (float) Correlation of elements in a block
        :return: (np.array) Resulting correlation matrix
        """

        # Creating a single block with all elements as block_corr
        block = np.ones((block_size, block_size)) * block_corr

        # Setting the main diagonal to ones
        block[range(block_size), range(block_size)] = 1

        # Create a block diagonal matrix with a number of equal blocks
        res_matrix = block_diag(*([block] * num_blocks))

        return res_matrix

    def form_true_matrix(self, num_blocks, block_size, block_corr, std=None):
        """
        Creates a random vector of means and a random covariance matrix.

        Due to the block structure of a matrix, it is a good sample data to use in the NCO and MCOS algorithms.

        The number of assets in a portfolio, number of blocks and correlations
        both inside the cluster and between clusters are adjustable.

        :param num_blocks: (int) Number of blocks in matrix
        :param block_size: (int) Size of a single block
        :param block_corr: (float) Correlation of elements in a block
        :param std: (float) Correlation between the clusters. If None, taken a random value from uniform dist[0.05, 0.2]
        :return: (np.array, pd.DataFrame) Resulting vector of means and the dataframe with covariance matrix
        """

        # Creating a block correlation matrix
        corr_matrix = self._form_block_matrix(num_blocks, block_size, block_corr)

        # Transforming to DataFrame
        corr_matrix = pd.DataFrame(corr_matrix)

        # Getting columns of matrix separately
        columns = corr_matrix.columns.tolist()

        # Randomizing the order of the columns
        np.random.shuffle(columns)
        corr_matrix = corr_matrix[columns].loc[columns].copy(deep=True)

        if std is None:  # Default intra-cluster correlations at 0.5
            std = np.random.uniform(.05, .2, corr_matrix.shape[0])
        else:  # Or the ones set by user
            std = np.array([std] * corr_matrix.shape[1])

        # Class to calculate covariance from the correlation function
        risk_estimators = RiskEstimators()

        # Calculating covariance matrix from the generated correlation matrix
        cov_matrix = risk_estimators.corr_to_cov(corr_matrix, std)

        # Vector of means
        mu_vec = np.random.normal(std, std, cov_matrix.shape[0]).reshape(-1, 1)

        return mu_vec, cov_matrix



"""
Logic regarding the various sampling techniques, in particular:

* Time Bars
* Tick Bars
* Volume Bars
* Dollar Bars
* Tick Imbalance Bars (EMA and Const)
* Volume Imbalance Bars (EMA and Const)
* Dollar Imbalance Bars (EMA and Const)
* Tick Run Bars (EMA and Const)
* Volume Run Bars (EMA and Const)
* Dollar Run Bars (EMA and Const)
"""

from famlafl.data_structures.imbalance_data_structures import get_ema_dollar_imbalance_bars, \
    get_ema_volume_imbalance_bars, \
    get_ema_tick_imbalance_bars, get_const_dollar_imbalance_bars, get_const_volume_imbalance_bars, \
    get_const_tick_imbalance_bars
from famlafl.data_structures.run_data_structures import get_ema_volume_run_bars, get_ema_tick_run_bars, \
    get_ema_dollar_run_bars, get_const_volume_run_bars, get_const_tick_run_bars, get_const_dollar_run_bars
from famlafl.data_structures.standard_data_structures import get_tick_bars, get_dollar_bars, get_volume_bars
from famlafl.data_structures.time_data_structures import get_time_bars



"""
Implementation of an algorithm described in Yimou Li, David Turkington, Alireza Yazdani
'Beyond the Black Box: An Intuitive Approach to Investment Prediction with Machine Learning'
(https://jfds.pm-research.com/content/early/2019/12/11/jfds.2019.1.023)
"""

from abc import ABC, abstractmethod
from typing import Tuple
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.linear_model import LinearRegression


# pylint: disable=invalid-name
# pylint: disable=too-many-locals

class AbstractModelFingerprint(ABC):
    """
    Model fingerprint constructor.

    This is an abstract base class for the RegressionModelFingerprint and ClassificationModelFingerprint classes.
    """

    def __init__(self):
        """
        Model fingerprint constructor.
        """

        # Effects containers
        self.linear_effect = None
        self.non_linear_effect = None
        self.pair_wise_effect = None

        self.ind_partial_dep_functions = None  # partial dependence function values containers
        self.feature_column_position_mapping = None  # numerical mapping between columns string and its position
        self.feature_values = None  # Feature values used in fingerprints analysis

    def fit(self, model: object, X: pd.DataFrame, num_values: int = 50, pairwise_combinations: list = None) -> None:
        """
        Get linear, non-linear and pairwise effects estimation.

        :param model: (object) Trained model.
        :param X: (pd.DataFrame) Dataframe of features.
        :param num_values: (int) Number of values used to estimate feature effect.
        :param pairwise_combinations: (list) Tuples (feature_i, feature_j) to test pairwise effect.
        """

        # Step 1
        self._get_feature_values(X, num_values)

        # Get partial dependency functions
        self._get_individual_partial_dependence(model, X)

        linear_effect = self._get_linear_effect(X)
        non_linear_effect = self._get_non_linear_effect(X)

        if pairwise_combinations is not None:
            pairwise_effect = self._get_pairwise_effect(pairwise_combinations, model, X, num_values)
            self.pair_wise_effect = {'raw': pairwise_effect, 'norm': self._normalize(pairwise_effect)}

        # Save results
        self.linear_effect = {'raw': linear_effect, 'norm': self._normalize(linear_effect)}
        self.non_linear_effect = {'raw': non_linear_effect, 'norm': self._normalize(non_linear_effect)}

    def get_effects(self) -> Tuple:
        """
        Return computed linear, non-linear and pairwise effects. The model should be fit() before using this method.

        :return: (tuple) Linear, non-linear and pairwise effects, of type dictionary (raw values and normalised).
        """
        return self.linear_effect, self.non_linear_effect, self.pair_wise_effect

    def plot_effects(self) -> plt.figure:
        """
        Plot each effect (normalized) on a bar plot (linear, non-linear). Also plots pairwise effects if calculated.

        :return: (plt.figure) Plot figure.
        """
        if self.pair_wise_effect is None:
            fig, (ax1, ax2) = plt.subplots(2, 1)
        else:
            fig, (ax1, ax2, ax3) = plt.subplots(3, 1)
            ax3.set_title('Pair-wise effect')
            ax3.bar(*zip(*self.pair_wise_effect['norm'].items()))

        ax1.set_title('Linear effect')
        ax1.bar(*zip(*self.linear_effect['norm'].items()))

        ax2.set_title('Non-Linear effect')
        ax2.bar(*zip(*self.non_linear_effect['norm'].items()))

        fig.tight_layout()
        return fig

    def _get_feature_values(self, X: pd.DataFrame, num_values: int) -> None:
        """
        Step 1 of the algorithm which generates possible feature values used in analysis.

        :param X: (pd.DataFrame) Dataframe of features.
        :param num_values: (int) Number of values used to estimate feature effect.
        """

        # Get possible feature values (Step 1)
        self.feature_values = {}  # Dictionary of feature values range used to build dependence functions
        for feature in X.columns:
            values = []
            for q in np.linspace(0, 1, num_values):
                values.append(np.quantile(X[feature], q=q))
            self.feature_values[feature] = np.array(values)

        # DataFrame with partial dependence function values for columns from X (f_k)
        self.ind_partial_dep_functions = pd.DataFrame(index=list(range(num_values)), columns=X.columns)

        self.feature_column_position_mapping = dict(zip(X.columns, range(0, X.shape[1])))

    def _get_individual_partial_dependence(self, model: object, X: pd.DataFrame) -> None:
        """
        Get individual partial dependence function values for each column.

        :param model: (object) Trained model.
        :param X: (pd.DataFrame) Dataframe of features.
        """
        for col in X.columns:
            y_mean_arr = []  # Array of mean(prediction) for each feature value
            for x_k in self.feature_values[col]:
                # Step 2
                col_k_position = self.feature_column_position_mapping[col]
                X_ = X.values.copy()
                X_[:, col_k_position] = x_k

                # Step 3
                y_pred = self._get_model_predictions(model, X_)
                y_pred_mean = np.mean(y_pred)

                y_mean_arr.append(y_pred_mean)

            self.ind_partial_dep_functions[col] = y_mean_arr

    def _get_linear_effect(self, X: pd.DataFrame) -> dict:
        """
        Get linear effect estimates as the mean absolute deviation of the linear predictions around their average value.

        :param X: (pd.DataFrame) Dataframe of features.
        :return: (dict) Linear effect estimates for each feature column.
        """
        store = {}
        for col in X.columns:
            x = self.feature_values[col].reshape(-1, 1)
            y = self.ind_partial_dep_functions[col]

            lmodel = LinearRegression(fit_intercept=True)
            lmodel.fit(x, y)
            y_mean = np.mean(y)
            linear_effect = np.mean(np.abs(lmodel.predict(x) - y_mean))
            store[col] = linear_effect
        return store

    def _get_non_linear_effect(self, X: pd.DataFrame) -> dict:
        """
        Get non-linear effect estimates as as the mean absolute deviation of the total marginal (single variable)
        effect around its corresponding linear effect.

        :param X: (pd.DataFrame) Dataframe of features.
        :return: (dict) Non-linear effect estimates for each feature column.
        """
        store = {}
        for col in X.columns:
            x = self.feature_values[col].reshape(-1, 1)
            y = self.ind_partial_dep_functions[col]

            lmodel = LinearRegression(fit_intercept=True)
            lmodel.fit(x, y)
            nonlinear_effect = np.mean(np.abs(lmodel.predict(x) - y.values))
            store[col] = nonlinear_effect
        return store

    def _get_pairwise_effect(self, pairwise_combinations: list, model: object, X: pd.DataFrame, num_values) -> dict:
        """
        Get pairwise effect estimates as the de-meaned joint partial prediction of the two variables minus the de-meaned
        partial predictions of each variable independently.

        :param pairwise_combinations: (list) Tuples (feature_i, feature_j) to test pairwise effect.
        :param model: (object) Trained model.
        :param X: (pd.DataFrame) Dataframe of features.
        :param num_values: (int) Number of values used to estimate feature effect.
        :return: (dict) Raw and normalised pairwise effects.
        """
        store = {}

        for pair in pairwise_combinations:
            function_values = []  # Array of pairwise interactions [f_k_l(not centered), f_k(centered), f_l(centered)]
            col_k = pair[0]
            col_l = pair[1]

            y_cdf_k_centered = self.ind_partial_dep_functions[col_k] - np.mean(self.ind_partial_dep_functions[col_k])
            y_cdf_l_centered = self.ind_partial_dep_functions[col_l] - np.mean(self.ind_partial_dep_functions[col_l])

            for x_k, y_cdf_k in zip(self.feature_values[col_k], y_cdf_k_centered):
                for x_l, y_cdf_l in zip(self.feature_values[col_l], y_cdf_l_centered):
                    col_k_position = self.feature_column_position_mapping[col_k]
                    col_l_position = self.feature_column_position_mapping[col_l]
                    X_ = X.values.copy()
                    X_[:, col_k_position] = x_k
                    X_[:, col_l_position] = x_l

                    y_cdf_k_l = self._get_model_predictions(model, X_).mean()

                    function_values.append([y_cdf_k_l, y_cdf_k, y_cdf_l])

            function_values = np.array(function_values)  # Convert to np.array to vectorize operations

            # Cumulative pairwise effect value for a given feature
            # Need to center f_k_l as f_k, f_l have been already centered
            centered_y_cdf_k_l = function_values[:, 0] - np.mean(function_values[:, 0])

            # See the paper to make use of notation
            f_k = function_values[:, 1]
            f_l = function_values[:, 2]
            func_value = sum(abs((centered_y_cdf_k_l - f_k - f_l)))

            store[str(pair)] = func_value / (num_values ** 2)

        return store

    @abstractmethod
    def _get_model_predictions(self, model: object, X_: pd.DataFrame):
        """
        Get model predictions based on problem type (predict for regression, predict_proba for classification).

        :param model: (object) Trained model.
        :param X_: (np.array) Feature set.
        :return: (np.array) Predictions.
        """
        raise NotImplementedError('Must implement _get_model_predictions')

    @staticmethod
    def _normalize(effect: dict) -> dict:
        """
        Normalize effect values (sum equals 1).

        :param effect: (dict) Effect values.
        :return: (dict) Normalized effect values.
        """
        values_sum = sum(effect.values())
        updated_effect = {}

        for k, v in effect.items():
            updated_effect[k] = v / values_sum
        return updated_effect


class RegressionModelFingerprint(AbstractModelFingerprint):
    """
    Regression Fingerprint class used for regression type of models.
    """

    def __init__(self):
        """
        Regression model fingerprint constructor.
        """
        AbstractModelFingerprint.__init__(self)

    def _get_model_predictions(self, model, X_):
        """
        Abstract method _get_model_predictions implementation.

        :param model: (object) Trained model.
        :param X_: (np.array) Feature set.
        :return: (np.array) Predictions.
        """
        return model.predict(X_)


class ClassificationModelFingerprint(AbstractModelFingerprint):
    """
    Classification Fingerprint class used for classification type of models.
    """

    def __init__(self):
        """
        Classification model fingerprint constructor.
        """
        AbstractModelFingerprint.__init__(self)

    def _get_model_predictions(self, model, X_):
        """
        Abstract method _get_model_predictions implementation.

        :param model: (object) Trained model.
        :param X_: (np.array) Feature set.
        :return: (np.array) Predictions.
        """
        return model.predict_proba(X_)[:, 1]





# pylint: disable=missing-module-docstring
class ReturnsEstimators:
    """
    This class contains methods for estimating expected returns. A good estimation of the asset expected returns is very important
    for portfolio optimisation problems and so it is necessary to use good estimates of returns and not just rely on
    simple techniques.
    """

    def __init__(self):
        """
        Initialize
        """

        return

    @staticmethod
    def calculate_mean_historical_returns(asset_prices, resample_by=None, frequency=252):
        """
        Calculates the annualised mean historical returns from asset price data.

        :param asset_prices: (pd.DataFrame) Asset price data
        :param resample_by: (str) Period to resample data ['D','W','M' etc.] None for no resampling
        :param frequency: (int) Average number of observations per year
        :return: (pd.Series) Annualized mean historical returns per asset
        """

        # Resample the asset prices
        if resample_by:
            asset_prices = asset_prices.resample(resample_by).last()
        returns = asset_prices.pct_change().dropna(how="all")
        returns = returns.mean() * frequency
        return returns

    @staticmethod
    def calculate_exponential_historical_returns(asset_prices, resample_by=None, frequency=252, span=500):
        """
        Calculates the exponentially-weighted annualized mean of historical returns, giving
        higher weight to more recent data.

        :param asset_prices: (pd.DataFrame) Asset price data
        :param resample_by: (str) Period to resample data ['D','W','M' etc.] None for no resampling
        :param frequency: (int) Average number of observations per year
        :param span: (int) Window length to use in pandas ewm function
        :return: (pd.Series) Exponentially-weighted mean of historical returns
        """

        # Resample the asset prices
        if resample_by:
            asset_prices = asset_prices.resample(resample_by).last()
        returns = asset_prices.pct_change().dropna(how="all")
        returns = returns.ewm(span=span).mean().iloc[-1] * frequency
        return returns

    @staticmethod
    def calculate_returns(asset_prices, resample_by=None):
        """
        Calculates a dataframe of returns from a dataframe of prices.

        :param asset_prices: (pd.DataFrame) Historical asset prices
        :param resample_by: (str) Period to resample data ['D','W','M' etc.] None for no resampling
        :return: (pd.DataFrame) Returns per asset
        """

        if resample_by:
            asset_prices = asset_prices.resample(resample_by).last()
        asset_returns = asset_prices.pct_change()
        asset_returns = asset_returns.dropna(how='all')
        return asset_returns



# pylint: disable=missing-module-docstring
# pylint: disable=no-self-use

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.labeling.raw_return import raw_return


class TestLabelingRawReturns(unittest.TestCase):
    """
    Tests for the raw returns labeling method.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(self.path, index_col='Date', parse_dates=True)
        self.idx5 = self.data[:5].index
        self.col5 = self.data.iloc[:, 0:5].columns

    def test_dataframe(self):
        """
        Verifies raw returns for a DataFrame.
        """
        prices = self.data.iloc[0:5, 0:5]
        test1 = raw_return(prices, lag=False)
        test2 = raw_return(prices, binary=True, lag=False)
        test3 = raw_return(prices, logarithmic=True, lag=True)
        test4 = raw_return(prices, binary=True, logarithmic=True, lag=True)
        test1_actual = pd.DataFrame([(np.nan, np.nan, np.nan, np.nan, np.nan),
                                     (0.008997, -0.002826, 0.0033758, 0.003779, 0.001662),
                                     (-0.030037, -0.019552, -0.0002804, -0.025602, -0.022719),
                                     (0.007327, 0.000867, -0.000187, -0.006182, 0.001045),
                                     (-0.007754, -0.006930, 0.000748, -0.002333, -0.005610)],
                                    columns=self.col5, index=self.idx5)

        test3_actual = pd.DataFrame([(0.008957, -0.002830, 0.003370, 0.003772, 0.001661),
                                     (-0.030497, -0.019746, -0.0002804, -0.025936, -0.022981),
                                     (0.007300, 0.000867, -0.000187, -0.006202, 0.001044),
                                     (-0.007784, -0.006955, 0.0007478, -0.002336, -0.005626),
                                     (np.nan, np.nan, np.nan, np.nan, np.nan)],
                                    columns=self.col5, index=self.idx5)

        pd.testing.assert_frame_equal(test1, test1_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test2, test1_actual.apply(np.sign))
        pd.testing.assert_frame_equal(test3, test3_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test4, test3_actual.apply(np.sign))

    def test_series(self):
        """
        Verifies raw returns for a series for simple/logarithmic returns, with numerical/binary labels.
        """
        # Takes series of 10 imaginary prices.
        price = pd.Series([100, 101, 102, 102, 102, 99, 19, 2000, 100, 105])

        test4 = raw_return(price, lag=True)
        test5 = raw_return(price, logarithmic=True, lag=False)
        test6 = raw_return(price, binary=True, logarithmic=True, lag=False)
        test4_actual = pd.Series([0.01, 0.009901, 0, 0, -0.029412, -0.808081, 104.263158, -0.95, 0.05, np.nan])
        test5_actual = pd.Series([np.nan, 0.00995033, 0.0098523, 0, 0, -0.02985296, -1.65068087, 4.65646348,
                                  -2.99573227, 0.04879016])
        pd.testing.assert_series_equal(test4, test4_actual, rtol=0.001)
        pd.testing.assert_series_equal(test5, test5_actual, rtol=0.001)
        pd.testing.assert_series_equal(test6, test5_actual.apply(np.sign))

    def test_resample(self):
        """
        Tests that resampling works correctly.
        """
        price1 = self.data.iloc[0:25, 0:5]
        price2 = self.data.iloc[:, 0:3]
        week_index = price1.resample('W').last().index
        year_index = self.data.resample('Y').last().index
        test6 = raw_return(price1, binary=False, logarithmic=True, resample_by='W', lag=True)
        test7 = raw_return(price2, binary=False, logarithmic=False, resample_by='Y', lag=True)
        test8 = raw_return(price2, binary=True, logarithmic=False, resample_by='Y', lag=True)
        test6_actual = pd.DataFrame([(0.014956, -0.014263, 0.002707, -0.012442, -0.018586),
                                     (-0.081178, -0.056693, 0.011494, -0.022153, -0.046989),
                                     (-0.015181, -0.071367, 0.006431, -0.002403, -0.014326),
                                     (0.045872, 0.046868, -0.001099, 0.028460, 0.039661),
                                     (-0.068778, -0.044871, -0.002203, -0.050350, -0.058079),
                                     (np.nan, np.nan, np.nan, np.nan, np.nan)],
                                    columns=self.col5, index=week_index)
        test7_actual = pd.DataFrame({'EEM': [0.661994, 0.147952, -0.203610, 0.168951, -0.057497, -0.060048, -0.180708,
                                             0.077664, np.nan],
                                     'EWG': [0.167534, 0.066845, -0.197160, 0.285120, 0.285830, -0.136965, -0.044509,
                                             -0.079038, np.nan],
                                     'TIP': [0.046957, 0.034841, 0.085287, 0.040449, -0.094803, 0.019199, -0.020802,
                                             0.067287, np.nan]}, index=year_index)
        pd.testing.assert_frame_equal(test6, test6_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test7, test7_actual, rtol=0.001)
        pd.testing.assert_frame_equal(test8, test7_actual.apply(np.sign))



"""
FAMLAFL helps portfolio managers and traders who want to leverage the power of machine learning by providing
reproducible, interpretable, and easy to use tools.

Adding FAMLAFL to your companies pipeline is like adding a department of PhD researchers to your team.
"""

import famlafl.util as util
import famlafl.data_structures as data_structures
import famlafl.datasets as datasets
import famlafl.multi_product as multi_product
import famlafl.filters.filters as filters
import famlafl.labeling as labeling
import famlafl.features.fracdiff as fracdiff
import famlafl.sample_weights as sample_weights
import famlafl.sampling as sampling
import famlafl.bet_sizing as bet_sizing
import famlafl.structural_breaks as structural_breaks
import famlafl.feature_importance as feature_importance
import famlafl.ensemble as ensemble
import famlafl.portfolio_optimization as portfolio_optimization
import famlafl.clustering as clustering
import famlafl.microstructural_features as microstructural_features
import famlafl.cross_validation as cross_validation
import famlafl.online_portfolio_selection as online_portfolio_selection
from famlafl.backtest_statistics import statistics
from famlafl.backtest_statistics import backtests



"""
Advances in Financial Machine Learning, Marcos Lopez de Prado
Chapter 2: Financial Data Structures: Imbalance Bars

This module contains the functions to help users create structured financial data from raw unstructured data,
in the form of tick, volume, and dollar imbalance bars.

These bars are used throughout the text book (Advances in Financial Machine Learning, By Marcos Lopez de Prado, 2018,
pg 29) to build the more interesting features for predicting financial time series data.

These financial data structures have better statistical properties when compared to those based on fixed time
interval sampling. A great paper to read more about this is titled: The Volume Clock: Insights into the high
frequency paradigm, Lopez de Prado, et al. These ideas are then extended in another paper: Flow toxicity and liquidity
in a high-frequency world.

We have introduced two types of imbalance bars: with expected number of tick defined through EMA (book implementation) and
constant number of ticks.

A good blog post to read, which helped us a lot in the implementation here is writen by Maksim Ivanov:
https://towardsdatascience.com/financial-machine-learning-part-0-bars-745897d4e4ba
"""

# Imports
from typing import Union, Iterable, List, Optional

import numpy as np
import pandas as pd

from famlafl.data_structures.base_bars import BaseImbalanceBars
from famlafl.util.fast_ewma import ewma


class EMAImbalanceBars(BaseImbalanceBars):
    """
    Contains all of the logic to construct the imbalance bars from chapter 2. This class shouldn't be used directly.
    We have added functions to the package such as get_ema_dollar_imbalance_bars which will create an instance of this
    class and then construct the imbalance bars, to return to the user.

    This is because we wanted to simplify the logic as much as possible, for the end user.
    """

    def __init__(self, metric: str, num_prev_bars: int,
                 expected_imbalance_window: int, exp_num_ticks_init: int,
                 exp_num_ticks_constraints: List, batch_size: int, analyse_thresholds: bool):
        """
        Constructor

        :param metric: (str) Type of imbalance bar to create. Example: "dollar_imbalance"
        :param num_prev_bars: (int) Window size for E[T]s (number of previous bars to use for expected number of ticks estimation)
        :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
        :param exp_num_ticks_init: (int) Initial number of expected ticks
        :param exp_num_ticks_constraints (list) Minimum and maximum possible number of expected ticks. Used to control bars sampling convergence
        :param batch_size: (int) Number of rows to read in from the csv, per batch
        :param analyse_thresholds: (bool) flag to return thresholds values (theta, exp_num_ticks, exp_imbalance) in a
                                          form of Pandas DataFrame
        """
        BaseImbalanceBars.__init__(self, metric, batch_size, expected_imbalance_window,
                                   exp_num_ticks_init, analyse_thresholds)

        # EMA Imbalance specific  hyper parameters
        self.num_prev_bars = num_prev_bars
        if exp_num_ticks_constraints is None:
            self.min_exp_num_ticks = 0
            self.max_exp_num_ticks = np.inf
        else:
            self.min_exp_num_ticks = exp_num_ticks_constraints[0]
            self.max_exp_num_ticks = exp_num_ticks_constraints[1]

    def _get_exp_num_ticks(self):
        prev_num_of_ticks = self.imbalance_tick_statistics['num_ticks_bar']
        exp_num_ticks = ewma(np.array(
            prev_num_of_ticks[-self.num_prev_bars:], dtype=float), self.num_prev_bars)[-1]
        return min(max(exp_num_ticks, self.min_exp_num_ticks), self.max_exp_num_ticks)


class ConstImbalanceBars(BaseImbalanceBars):
    """
    Contains all of the logic to construct the imbalance bars with fixed expected number of ticks. This class shouldn't
    be used directly. We have added functions to the package such as get_const_dollar_imbalance_bars which will create
    an instance of this class and then construct the imbalance bars, to return to the user.

    This is because we wanted to simplify the logic as much as possible, for the end user.
    """

    def __init__(self, metric: str, expected_imbalance_window: int,
                 exp_num_ticks_init: int, batch_size: int,
                 analyse_thresholds: bool):
        """
        Constructor

        :param metric: (str) Type of imbalance bar to create. Example: "dollar_imbalance"
        :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
        :param exp_num_ticks_init: (int) Initial number of expected ticks
        :param batch_size: (int) Number of rows to read in from the csv, per batch
        :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
        """
        BaseImbalanceBars.__init__(self, metric, batch_size, expected_imbalance_window,
                                   exp_num_ticks_init,
                                   analyse_thresholds)

    def _get_exp_num_ticks(self):
        return self.thresholds['exp_num_ticks']


def get_ema_dollar_imbalance_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], num_prev_bars: int = 3,
                                  expected_imbalance_window: int = 10000, exp_num_ticks_init: int = 20000,
                                  exp_num_ticks_constraints: List[float] = None, batch_size: int = 2e7,
                                  analyse_thresholds: bool = False,
                                  verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the EMA dollar imbalance bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param num_prev_bars: (int) Window size for E[T]s (number of previous bars to use for expected number of ticks estimation)
    :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
    :param exp_num_ticks_init: (int) Initial expected number of ticks per bar
    :param exp_num_ticks_constraints: (list) Minimum and maximum possible number of expected ticks. Used to control bars sampling convergence
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) DataFrame of dollar imbalance bars and DataFrame of thresholds, if to_csv=True returns None
    """
    bars = EMAImbalanceBars(metric='dollar_imbalance', num_prev_bars=num_prev_bars,
                            expected_imbalance_window=expected_imbalance_window,
                            exp_num_ticks_init=exp_num_ticks_init, exp_num_ticks_constraints=exp_num_ticks_constraints,
                            batch_size=batch_size, analyse_thresholds=analyse_thresholds)
    imbalance_bars = bars.batch_run(file_path_or_df=file_path_or_df,
                                    verbose=verbose, to_csv=to_csv, output_path=output_path)

    return imbalance_bars, pd.DataFrame(bars.bars_thresholds)


def get_ema_volume_imbalance_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], num_prev_bars: int = 3,
                                  expected_imbalance_window: int = 10000, exp_num_ticks_init: int = 20000,
                                  exp_num_ticks_constraints: List[float] = None, batch_size: int = 2e7,
                                  analyse_thresholds: bool = False,
                                  verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the EMA volume imbalance bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param num_prev_bars: (int) Window size for E[T]s (number of previous bars to use for expected number of ticks estimation)
    :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
    :param exp_num_ticks_init: (int) Initial expected number of ticks per bar
    :param exp_num_ticks_constraints: (list) Minimum and maximum possible number of expected ticks. Used to control bars sampling convergence
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) DataFrame of volume imbalance bars and DataFrame of thresholds, if to_csv=True returns None
    """
    bars = EMAImbalanceBars(metric='volume_imbalance', num_prev_bars=num_prev_bars,
                            expected_imbalance_window=expected_imbalance_window,
                            exp_num_ticks_init=exp_num_ticks_init, exp_num_ticks_constraints=exp_num_ticks_constraints,
                            batch_size=batch_size, analyse_thresholds=analyse_thresholds)
    imbalance_bars = bars.batch_run(file_path_or_df=file_path_or_df,
                                    verbose=verbose, to_csv=to_csv, output_path=output_path)

    return imbalance_bars, pd.DataFrame(bars.bars_thresholds)


def get_ema_tick_imbalance_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], num_prev_bars: int = 3,
                                expected_imbalance_window: int = 10000, exp_num_ticks_init: int = 20000,
                                exp_num_ticks_constraints: List[float] = None, batch_size: int = 2e7,
                                analyse_thresholds: bool = False,
                                verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the EMA tick imbalance bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                             in the format[date_time, price, volume]
    :param num_prev_bars: (int) Window size for E[T]s (number of previous bars to use for expected number of ticks estimation)
    :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
    :param exp_num_ticks_init: (int) Initial expected number of ticks per bar
    :param exp_num_ticks_constraints: (array) Minimum and maximum possible number of expected ticks. Used to control bars sampling convergence
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) DataFrame of tick imbalance bars and DataFrame of thresholds, if to_csv=True returns None
    """
    bars = EMAImbalanceBars(metric='tick_imbalance', num_prev_bars=num_prev_bars,
                            expected_imbalance_window=expected_imbalance_window,
                            exp_num_ticks_init=exp_num_ticks_init, exp_num_ticks_constraints=exp_num_ticks_constraints,
                            batch_size=batch_size, analyse_thresholds=analyse_thresholds)
    imbalance_bars = bars.batch_run(file_path_or_df=file_path_or_df,
                                    verbose=verbose, to_csv=to_csv, output_path=output_path)

    return imbalance_bars, pd.DataFrame(bars.bars_thresholds)


def get_const_dollar_imbalance_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], expected_imbalance_window: int = 10000,
                                    exp_num_ticks_init: int = 20000,
                                    batch_size: int = 2e7, analyse_thresholds: bool = False,
                                    verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the Const dollar imbalance bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
    :param exp_num_ticks_init: (int) Initial expected number of ticks per bar
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) DataFrame of dollar imbalance bars and DataFrame of thresholds, if to_csv=True returns None
    """
    bars = ConstImbalanceBars(metric='dollar_imbalance',
                              expected_imbalance_window=expected_imbalance_window,
                              exp_num_ticks_init=exp_num_ticks_init,
                              batch_size=batch_size, analyse_thresholds=analyse_thresholds)
    imbalance_bars = bars.batch_run(file_path_or_df=file_path_or_df,
                                    verbose=verbose, to_csv=to_csv, output_path=output_path)

    return imbalance_bars, pd.DataFrame(bars.bars_thresholds)


def get_const_volume_imbalance_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], expected_imbalance_window: int = 10000,
                                    exp_num_ticks_init: int = 20000,
                                    batch_size: int = 2e7, analyse_thresholds: bool = False,
                                    verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the Const volume imbalance bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
    :param exp_num_ticks_init: (int) Initial expected number of ticks per bar
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) DataFrame of volume imbalance bars and DataFrame of thresholds, if to_csv=True returns None
    """
    bars = ConstImbalanceBars(metric='volume_imbalance',
                              expected_imbalance_window=expected_imbalance_window,
                              exp_num_ticks_init=exp_num_ticks_init,
                              batch_size=batch_size, analyse_thresholds=analyse_thresholds)
    imbalance_bars = bars.batch_run(file_path_or_df=file_path_or_df,
                                    verbose=verbose, to_csv=to_csv, output_path=output_path)

    return imbalance_bars, pd.DataFrame(bars.bars_thresholds)


def get_const_tick_imbalance_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], expected_imbalance_window: int = 10000,
                                  exp_num_ticks_init: int = 20000,
                                  batch_size: int = 2e7, analyse_thresholds: bool = False,
                                  verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates the Const tick imbalance bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str or pd.DataFrame) Path to the csv file or Pandas Data Frame containing raw tick data in the format[date_time, price, volume]
    :param expected_imbalance_window: (int) EMA window used to estimate expected imbalance
    :param exp_num_ticks_init: (int) Initial expected number of ticks per bar
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (bool) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param analyse_thresholds: (bool) Flag to save  and return thresholds used to sample imbalance bars
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) DataFrame of tick imbalance bars and DataFrame of thresholds, if to_csv=True returns None
    """
    bars = ConstImbalanceBars(metric='tick_imbalance',
                              expected_imbalance_window=expected_imbalance_window,
                              exp_num_ticks_init=exp_num_ticks_init,
                              batch_size=batch_size, analyse_thresholds=analyse_thresholds)
    imbalance_bars = bars.batch_run(file_path_or_df=file_path_or_df,
                                    verbose=verbose, to_csv=to_csv, output_path=output_path)

    return imbalance_bars, pd.DataFrame(bars.bars_thresholds)



"""
Initialize Pattern Matching Module.
"""
from famlafl.online_portfolio_selection.corn import CORN
from famlafl.online_portfolio_selection.cornu import CORNU
from famlafl.online_portfolio_selection.cornk import CORNK
from famlafl.online_portfolio_selection.scorn import SCORN
from famlafl.online_portfolio_selection.scornk import SCORNK
from famlafl.online_portfolio_selection.fcorn import FCORN
from famlafl.online_portfolio_selection.fcornk import FCORNK



"""
Initialize Mean Reversion Module.
"""
from famlafl.online_portfolio_selection.rmr import RMR
from famlafl.online_portfolio_selection.pamr import PAMR
from famlafl.online_portfolio_selection.cwmr import CWMR
from famlafl.online_portfolio_selection.olmar import OLMAR



# pylint: disable=missing-module-docstring
# pylint: disable=protected-access

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.labeling.tail_sets import TailSetLabels


class TestTailSets(unittest.TestCase):
    """
    Unit tests for the tail sets labeling class.
    """

    def setUp(self):
        """
        Set the file path for the sample data.
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(self.path, index_col='Date', parse_dates=True)

    def test_init(self):
        """
        Tests to ensure exceptions are correctly raised for invalid inputs.
        """
        # Too large n_bins.
        with self.assertRaises(Exception):
            TailSetLabels(self.data[:100], n_bins=200)
        # Window isn't int with vol_adj.
        with self.assertRaises(Exception):
            TailSetLabels(self.data[:100], n_bins=10, vol_adj='stdev', window='str')
        # Window is too large
            TailSetLabels(self.data[:100], n_bins=10, vol_adj='stdev', window=200)

    def test_vol_adj_ret(self):
        """
        Tests for getting the correct volatility adjusted return.
        """
        # Check for no vol adjustment. Should be same as straight log returns.
        label1 = TailSetLabels(self.data[:100], n_bins=23, vol_adj=None, window=None)  # No vol adjustment
        test1 = label1.vol_adj_rets
        simple_returns = np.log(self.data[:100]).diff().dropna()

        # # Stdev adjusted returns.
        label2 = TailSetLabels(self.data[:100], n_bins=8, vol_adj='stdev', window=20)
        test2 = label2.vol_adj_rets
        test2_actual = (simple_returns / simple_returns.rolling(20).std()).dropna()

        # Mean abs returns
        label3 = TailSetLabels(self.data[:100], n_bins=12, vol_adj='mean_abs_dev', window=20)
        test3 = label3.vol_adj_rets
        pd.testing.assert_frame_equal(test1, simple_returns)
        pd.testing.assert_frame_equal(test2, test2_actual)
        np.testing.assert_array_almost_equal(test3.iloc[4:5, 3:13],
                                             np.array([[-0.88220253, -0.05699642, -0.41151834, 0.22209753, -0.26852039,
                                                        -0.41058931, -0.72457246, -0.45304492, -1.62358571,
                                                        -0.74637241]]))

    def test_extract_tail_sets(self):
        """
        Tests for extracting the tail set in one row, including positive and negative class.
        """
        label4 = TailSetLabels(self.data[:100], n_bins=4, vol_adj='mean_abs_dev', window=20)
        returns4 = label4.vol_adj_rets
        test4 = label4._extract_tail_sets(row=returns4.iloc[25])
        test4_actual = pd.Series([0, 0, 1, -1, 0, 1, 0, 0, -1, -1, 1, 0, 0, 0, -1, 0, 0, -1, 0, 1, 1, 1, -1],
                                 index=returns4.iloc[25].index)
        test5 = label4._positive_tail_set(test4)
        test5_actual = ['TIP', 'IEF', 'XLF', 'TLT', 'BND', 'CSJ']
        test6 = label4._negative_tail_set(test4)
        test6_actual = ['EWJ', 'XLB', 'XLE', 'EPP', 'VPL', 'DIA']
        pd.testing.assert_series_equal(test4, test4_actual)
        self.assertEqual(test5, test5_actual)
        self.assertEqual(test6, test6_actual)

    def test_overall(self):
        """
        Tests the overall output of the tail set labels.
        """
        label7 = TailSetLabels(self.data[:100], n_bins=10, vol_adj='mean_abs_dev', window=30)
        test7_pos, test7_neg, _ = label7.get_tail_sets()

        self.assertEqual(test7_pos[1], ['XLU', 'EPP', 'FXI'])
        self.assertEqual(test7_neg[1], ['EWU', 'XLK', 'DIA'])



"""
Tests Functional Correlation Driven Nonparametric Learning.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.fcorn \
    import FCORN


class TestFunctionalCorrelationDrivenNonparametricLearning(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Functional Correlation Driven Nonparametric Learning class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_fcorn_solution(self):
        """
        Test the calculation of FCORN.
        """
        # Initialize FCORN.
        fcorn = FCORN(window=2, rho=0.5, lambd=10)
        # Allocates asset prices to FCORN.
        fcorn.allocate(self.data, resample_by='3M')
        # Create np.array of all_weights.
        all_weights = np.array(fcorn.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_fcorn_window_error(self):
        """
        Tests ValueError if window is not an integer or less than 1.
        """
        # Initialize FCORN.
        fcorn1 = FCORN(window=2.5, rho=0.5, lambd=1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            fcorn1.allocate(self.data)

        # Initialize FCORN.
        fcorn2 = FCORN(window=0, rho=0.5, lambd=2)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            fcorn2.allocate(self.data)

    def test_fcorn_rho_error(self):
        """
        Tests ValueError if rho is less than -1 or more than 1.
        """
        # Initialize FCORN.
        fcorn3 = FCORN(window=2, rho=-2, lambd=4)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            fcorn3.allocate(self.data)

        # Initialize FCORN.
        fcorn4 = FCORN(window=2, rho=2, lambd=8)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            fcorn4.allocate(self.data)

    def test_sigmoid(self):
        # pylint: disable=no-self-use
        # pylint: disable=protected-access
        """
        Tests Sigmoid Calculation.
        """
        # Initialize FCORN.
        fcorn5 = FCORN(window=1, rho=0.5, lambd=10)
        sig = fcorn5._sigmoid(0.0)
        np.testing.assert_almost_equal(sig, 0.5)

    def test_fcorn_solution1(self):
        """
        Test the calculation of FCORN for edge case that activation function is 0.
        """
        # Initialize FCORN.
        fcorn6 = FCORN(window=1, rho=0.5, lambd=10)
        # Allocates asset prices to FCORN.
        fcorn6.allocate(self.data, resample_by='Y')
        # Create np.array of all_weights.
        all_weights = np.array(fcorn6.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)



"""
Functions derived from Chapter 7: Cross Validation
"""

from famlafl.cross_validation.cross_validation import (
    ml_get_train_times,
    ml_cross_val_score,
    PurgedKFold
)

from famlafl.cross_validation.combinatorial import CombinatorialPurgedKFold

__all__ = [
    'ml_get_train_times',
    'ml_cross_val_score',
    "PurgedKFold"
]



"""
Test various functions regarding chapter 8: MDI, MDA, SFI importance.
"""
import os
import unittest

import numpy as np

from sklearn.ensemble import RandomForestClassifier, BaggingClassifier
from sklearn.model_selection import cross_val_score, KFold
from sklearn.metrics import f1_score, log_loss
from famlafl.feature_importance.importance import (mean_decrease_impurity,
                                                    mean_decrease_accuracy, single_feature_importance,
                                                    plot_feature_importance)
from famlafl.feature_importance.orthogonal import feature_pca_analysis, get_orthogonal_features
from famlafl.clustering.feature_clusters import get_feature_clusters
from famlafl.util.generate_dataset import get_classification_data


# pylint: disable=invalid-name
class TestFeatureImportance(unittest.TestCase):
    """
    Test Feature importance
    """

    def setUp(self):
        """
        Generate X, y data sets and fit a RF
        """
        # Generate data sets
        self.X, self.y = get_classification_data(10, 5, 2, 1000, random_state=0, sigma=0)
        # Fit a RF
        self.clf_base = RandomForestClassifier(n_estimators=1, criterion='entropy', bootstrap=False,
                                               class_weight='balanced_subsample')

        self.bag_clf = BaggingClassifier(estimator=self.clf_base, max_features=1.0, n_estimators=100,
                                         oob_score=True, random_state=1)
        self.fit_clf = self.bag_clf.fit(self.X, self.y)
        self.cv_gen = KFold(n_splits=3)

    def test_orthogonal_features(self):
        """
        Test orthogonal features: PCA features, importance vs PCA importance analysis
        """

        pca_features = get_orthogonal_features(self.X)

        # PCA features should have mean of 0
        self.assertAlmostEqual(np.mean(pca_features[:, 2]), 0, delta=1e-7)
        self.assertAlmostEqual(np.mean(pca_features[:, 5]), 0, delta=1e-7)
        self.assertAlmostEqual(np.mean(pca_features[:, 6]), 0, delta=1e-7)

        # Check particular PCA values std
        self.assertAlmostEqual(np.std(pca_features[:, 1]), 1.3813, delta=0.2)
        self.assertAlmostEqual(np.std(pca_features[:, 3]), 1.0255, delta=0.2)
        self.assertAlmostEqual(np.std(pca_features[:, 4]), 1.0011, delta=0.2)

        mdi_feat_imp = mean_decrease_impurity(self.fit_clf, self.X.columns)
        pca_corr_res = feature_pca_analysis(self.X, mdi_feat_imp)

        # Check correlation metrics results
        self.assertAlmostEqual(pca_corr_res['Weighted_Kendall_Rank'][0], 0.7424, delta=1e-1)

        # Check particular number of PCA features
        pca_ten_features = get_orthogonal_features(self.X, num_features=10)
        self.assertEqual(pca_ten_features.shape[1], 10)

        pca_five_features = get_orthogonal_features(self.X, num_features=5)
        self.assertEqual(pca_five_features.shape[1], 5)

    def test_feature_importance(self):
        """
        Test features importance: MDI, MDA, SFI and plot function
        """
        # Getting the clustered subsets for CFI with number of clusters selection using ONC algorithm
        clustered_subsets_linear = get_feature_clusters(self.X, dependence_metric='linear',
                                                        distance_metric=None, linkage_method=None,
                                                        n_clusters=None)
        # Also to verify the theory that if number clusters is equal to number of features then the
        # result will be same as MDA
        feature_subset_single = [[x] for x in self.X.columns]

        # MDI feature importance
        mdi_feat_imp = mean_decrease_impurity(self.fit_clf, self.X.columns)
        # Clustered MDI feature importance
        clustered_mdi = mean_decrease_impurity(self.fit_clf, self.X.columns,
                                               clustered_subsets=clustered_subsets_linear)
        mdi_cfi_single = mean_decrease_impurity(self.fit_clf, self.X.columns,
                                                clustered_subsets=feature_subset_single)

        # MDA feature importance
        mda_feat_imp_log_loss = mean_decrease_accuracy(self.bag_clf, self.X, self.y, self.cv_gen,
                                                       sample_weight_train=np.ones((self.X.shape[0],)),
                                                       sample_weight_score=np.ones((self.X.shape[0],)),
                                                       scoring=log_loss)

        mda_feat_imp_f1 = mean_decrease_accuracy(self.bag_clf, self.X, self.y,
                                                 self.cv_gen, scoring=f1_score)
        # ClusteredMDA feature importance
        clustered_mda = mean_decrease_accuracy(self.bag_clf, self.X, self.y, self.cv_gen,
                                               clustered_subsets=clustered_subsets_linear)
        mda_cfi_single = mean_decrease_accuracy(self.bag_clf, self.X, self.y, self.cv_gen,
                                                clustered_subsets=feature_subset_single)

        # SFI feature importance
        sfi_feat_imp_log_loss = single_feature_importance(self.bag_clf, self.X,
                                                          self.y, cv_gen=self.cv_gen,
                                                          sample_weight_train=np.ones((self.X.shape[0],)),
                                                          scoring=log_loss)
        sfi_feat_imp_f1 = single_feature_importance(self.bag_clf, self.X,
                                                    self.y, cv_gen=self.cv_gen,
                                                    sample_weight_score=np.ones((self.X.shape[0],)),
                                                    scoring=f1_score)

        # MDI assertions
        self.assertAlmostEqual(mdi_feat_imp['mean'].sum(), 1, delta=0.001)
        # The most informative features
        # Value in the below test was changed after v.0.11.3 from 0.48058
        #                                                     to  0.46835
        # due to scikit-learn changing RandomForestClassifier, and BaggingClassifier outputs
        self.assertAlmostEqual(mdi_feat_imp.loc['I_1', 'mean'], 0.46835, delta=0.01)
        self.assertAlmostEqual(mdi_feat_imp.loc['I_0', 'mean'], 0.08214, delta=0.01)
        # Redundant feature
        self.assertAlmostEqual(mdi_feat_imp.loc['R_0', 'mean'], 0.06511, delta=0.01)
        # Noisy feature
        self.assertAlmostEqual(mdi_feat_imp.loc['N_0', 'mean'], 0.02229, delta=0.01)

        # MDA(log_loss) assertions
        self.assertAlmostEqual(mda_feat_imp_log_loss.loc['I_1', 'mean'], 0.65522, delta=0.1)
        # Value in the below test was changed after v.0.11.3 from 0.00332
        #                                                     to -0.20645
        # due to scikit-learn changing RandomForestClassifier, and BaggingClassifier outputs
        self.assertAlmostEqual(mda_feat_imp_log_loss.loc['R_0', 'mean'], -0.20645, delta=0.1)

        # MDA(f1) assertions
        self.assertAlmostEqual(mda_feat_imp_f1.loc['I_1', 'mean'], 0.47751, delta=0.1)
        self.assertAlmostEqual(mda_feat_imp_f1.loc['I_2', 'mean'], 0.33617, delta=0.1)

        # SFI(log_loss) assertions
        self.assertAlmostEqual(sfi_feat_imp_log_loss.loc['I_0', 'mean'], -6.39442, delta=0.1)
        self.assertAlmostEqual(sfi_feat_imp_log_loss.loc['R_0', 'mean'], -5.04315, delta=0.1)

        # SFI(accuracy) assertions
        self.assertAlmostEqual(sfi_feat_imp_f1.loc['I_0', 'mean'], 0.48915, delta=0.1)
        self.assertAlmostEqual(sfi_feat_imp_f1.loc['I_1', 'mean'], 0.78443, delta=0.1)

        # Cluster MDI  assertions
        self.assertAlmostEqual(clustered_mdi.loc['R_0', 'mean'], 0.01912, delta=0.1)
        self.assertAlmostEqual(clustered_mdi.loc['I_0', 'mean'], 0.06575, delta=0.1)

        # Clustered MDA (log_loss) assertions
        # Value in the below test was changed after v.0.11.3 from 0.04154
        #                                                     to -0.12248
        # due to scikit-learn changing RandomForestClassifier, and BaggingClassifier outputs
        self.assertAlmostEqual(clustered_mda.loc['I_0', 'mean'], -0.12248, delta=0.1)
        self.assertAlmostEqual(clustered_mda.loc['R_0', 'mean'], 0.02940, delta=0.1)

        # Test if CFI with number of clusters same to number features is equal to normal MDI & MDA results
        self.assertAlmostEqual(mdi_feat_imp.loc['I_1', 'mean'], mdi_cfi_single.loc['I_1', 'mean'], delta=0.1)
        self.assertAlmostEqual(mdi_feat_imp.loc['R_0', 'mean'], mdi_cfi_single.loc['R_0', 'mean'], delta=0.1)
        self.assertAlmostEqual(mda_feat_imp_log_loss.loc['I_1', 'mean'], mda_cfi_single.loc['I_1', 'mean'], delta=0.1)
        self.assertAlmostEqual(mda_feat_imp_log_loss.loc['R_0', 'mean'], mda_cfi_single.loc['R_0', 'mean'], delta=0.1)

    def test_plot_feature_importance(self):
        """
        Test plot_feature_importance function
        """
        oos_score = cross_val_score(self.bag_clf, self.X, self.y, cv=self.cv_gen, scoring='accuracy').mean()

        mdi_feat_imp = mean_decrease_impurity(self.bag_clf, self.X.columns)
        plot_feature_importance(mdi_feat_imp, oob_score=self.bag_clf.oob_score_, oos_score=oos_score)
        plot_feature_importance(mdi_feat_imp, oob_score=self.bag_clf.oob_score_, oos_score=oos_score,
                                save_fig=True, output_path='test.png')
        os.remove('test.png')
